from dataclasses import dataclass
from functools import cached_property
from typing import List, Optional, Dict, Union, Tuple
import numpy as np
from pathlib import Path

from pyrion.core.strand import Strand


@dataclass(frozen=True)
class GenomeAlignment:
    chain_id: int
    score: int
    t_chrom: str
    t_strand: int
    t_size: int
    q_chrom: str
    q_strand: int
    q_size: int
    blocks: np.ndarray             # Shape (N, 4) -- [[t_start, t_end, q_start, q_end], ...]
    child_id: Optional[int] = None  # For sub-chains (None for original chains)

    @cached_property
    def t_span(self) -> np.ndarray:
        return np.array([self.blocks[0][0], self.blocks[-1][1]], dtype=np.int32)

    @cached_property
    def q_span(self) -> np.ndarray:
        if self.q_strand == -1:
            # For negative strand, coordinates are reversed: first block has highest, last block has lowest
            return np.array([self.blocks[-1][2], self.blocks[0][3]], dtype=np.int32)
        else:
            # For positive strand, use first block start and last block end as before
            return np.array([self.blocks[0][2], self.blocks[-1][3]], dtype=np.int32)

    def __repr__(self) -> str:
        t_strand_str = Strand.from_int(self.t_strand).to_char()
        q_strand_str = Strand.from_int(self.q_strand).to_char()
        
        # Get target and query spans
        t_start, t_end = self.t_span
        q_start, q_end = self.q_span
        
        return (f"GenomeAlignment(id={self.chain_id}, score={self.score:,}, "
                f"T={self.t_chrom}:{t_start:,}-{t_end:,}:{t_strand_str} -> "
                f"Q={self.q_chrom}:{q_start:,}-{q_end:,}:{q_strand_str}, "
                f"{len(self.blocks)} blocks)")

    def aligned_length(self) -> int:
        if len(self.blocks) == 0:
            return 0
        return int(np.sum(self.blocks[:, 1] - self.blocks[:, 0]))

    def t_length(self) -> int:
        return int(self.t_span[1] - self.t_span[0])

    def q_length(self) -> int:
        return int(self.q_span[1] - self.q_span[0])

    def blocks_in_target(self) -> np.ndarray:
        return self.blocks[:, :2]

    def blocks_in_query(self) -> np.ndarray:
        return self.blocks[:, 2:]

    def __hash__(self):
        return hash((self.chain_id, self.t_chrom, self.t_strand, self.q_chrom, self.q_strand))


class GenomeAlignmentsCollection:
    """Container for many genome alignments."""
    def __init__(
        self,
        alignments: Optional[List[GenomeAlignment]] = None,
        source_file: Optional[str] = None
    ):
        self.alignments: List[GenomeAlignment] = alignments or []
        self.source_file: Optional[str] = source_file

        self._id_index: Optional[Dict[int, int]] = None  # chain_id → index
        self._chrom_index: Optional[Dict[str, List[int]]] = None  # target chrom → indices
        self._query_chrom_index: Optional[Dict[str, List[int]]] = None  # query chrom → indices

    def __len__(self):
        return len(self.alignments)

    def __getitem__(self, idx: int) -> GenomeAlignment:
        return self.alignments[idx]

    def get_by_chain_id(self, chain_id: int) -> Optional[GenomeAlignment]:
        if self._id_index is None:
            self._build_id_index()
        idx = self._id_index.get(chain_id)
        return self.alignments[idx] if idx is not None else None

    def get_by_target_chrom(self, chrom: str) -> List[GenomeAlignment]:
        if self._chrom_index is None:
            self._build_chrom_index()
        indices = self._chrom_index.get(chrom, [])
        return [self.alignments[i] for i in indices]

    def get_by_query_chrom(self, chrom: str) -> List[GenomeAlignment]:
        if self._query_chrom_index is None:
            self._build_query_chrom_index()
        indices = self._query_chrom_index.get(chrom, [])
        return [self.alignments[i] for i in indices]

    def get_reference_chromosomes(self) -> List[str]:
        if self._chrom_index is None:
            self._build_chrom_index()
        return list(self._chrom_index.keys())

    def get_query_chromosomes(self) -> List[str]:
        if self._query_chrom_index is None:
            self._build_query_chrom_index()
        return list(self._query_chrom_index.keys())

    def get_chain_ids_by_target_chrom(self, chrom: str) -> List[int]:
        if self._chrom_index is None:
            self._build_chrom_index()
        indices = self._chrom_index.get(chrom, [])
        return [self.alignments[i].chain_id for i in indices]

    def get_chain_ids_by_query_chrom(self, chrom: str) -> List[int]:
        if self._query_chrom_index is None:
            self._build_query_chrom_index()
        indices = self._query_chrom_index.get(chrom, [])
        return [self.alignments[i].chain_id for i in indices]

    def get_alignments_overlapping_target_interval(self, interval: 'GenomicInterval', include_partial: bool = True) -> List[GenomeAlignment]:
        target_chrom_alignments = self.get_by_target_chrom(interval.chrom)
        
        if not target_chrom_alignments:
            return []
        
        matching = []
        for alignment in target_chrom_alignments:
            # Get target span: [start, end]
            t_start, t_end = alignment.t_span
            
            if include_partial:
                if not (t_end <= interval.start or interval.end <= t_start):
                    matching.append(alignment)
            else:
                if t_start >= interval.start and t_end <= interval.end:
                    matching.append(alignment)
        
        return matching

    def get_alignments_overlapping_query_interval(self, interval: 'GenomicInterval', include_partial: bool = True) -> List[GenomeAlignment]:
        query_chrom_alignments = self.get_by_query_chrom(interval.chrom)
        
        if not query_chrom_alignments:
            return []
        
        # Check overlap on query coordinates
        matching = []
        for alignment in query_chrom_alignments:
            # Get query span: [start, end]  
            q_start, q_end = alignment.q_span
            
            if include_partial:
                # Include if alignment intersects (overlaps) the target interval - even 1nt
                if not (q_end <= interval.start or interval.end <= q_start):
                    matching.append(alignment)
            else:
                # Include only if alignment is fully contained within the target interval
                if q_start >= interval.start and q_end <= interval.end:
                    matching.append(alignment)
        
        return matching

    # Convenience methods with shorter names
    def get_alignments_in_interval(self, interval: 'GenomicInterval', include_partial: bool = True) -> List[GenomeAlignment]:
        return self.get_alignments_overlapping_target_interval(interval, include_partial)
    
    def get_alignments_fully_contained(self, interval: 'GenomicInterval') -> List[GenomeAlignment]:
        return self.get_alignments_overlapping_target_interval(interval, include_partial=False)

    def _build_id_index(self):
        self._id_index = {ga.chain_id: i for i, ga in enumerate(self.alignments)}

    def _build_chrom_index(self):
        from collections import defaultdict
        chrom_index = defaultdict(list)
        for i, ga in enumerate(self.alignments):
            chrom_index[ga.t_chrom].append(i)
        self._chrom_index = dict(chrom_index)

    def _build_query_chrom_index(self):
        from collections import defaultdict
        query_chrom_index = defaultdict(list)
        for i, ga in enumerate(self.alignments):
            query_chrom_index[ga.q_chrom].append(i)
        self._query_chrom_index = dict(query_chrom_index)

    def sort_by_score(self, max_elems: Optional[int] = None) -> List[Tuple[int, int]]:
        from .genome_alignment_auxiliary import sort_alignments_by_score
        return sort_alignments_by_score(self, max_elems)

    def summary(self) -> str:
        return f"{len(self.alignments):,} genome alignments from {self.source_file or 'unknown source'}"

    def __str__(self) -> str:
        """String representation with summary."""
        from ..ops.chain_serialization import genome_alignments_collection_summary_string
        return genome_alignments_collection_summary_string(self)

    def __repr__(self) -> str:
        return f"<GenomeAlignmentsCollection: {self.summary()}>"

    def save_to_chain(self, file_path: Union[str, Path]) -> None:
        from ..ops.chain_serialization import save_genome_alignments_collection_to_chain
        save_genome_alignments_collection_to_chain(self, file_path)

    def save_to_json(self, file_path: Union[str, Path]) -> None:
        from ..ops.chain_serialization import save_genome_alignments_collection_to_json
        save_genome_alignments_collection_to_json(self, file_path)

    @classmethod
    def from_json(cls, file_path: Union[str, Path]) -> 'GenomeAlignmentsCollection':
        from ..ops.chain_serialization import load_genome_alignments_collection_from_json
        return load_genome_alignments_collection_from_json(file_path)



"""Auxiliary functions for genome alignment objects."""

from typing import List, Tuple, Optional


def sort_alignments_by_score(alignments_collection, max_elems: Optional[int] = None) -> List[Tuple[int, int]]:
    score_tuples = [(alignment.chain_id, alignment.score) for alignment in alignments_collection.alignments]
    sorted_tuples = sorted(score_tuples, key=lambda x: x[1], reverse=True)
    
    if max_elems is not None:
        sorted_tuples = sorted_tuples[:max_elems]
    
    return sorted_tuples



"""Test GTF and BED12 parsing equivalence.

Compares transcripts parsed from GTF files with those parsed from equivalent BED12 files
to ensure consistency between parsing methods. Uses specific test files that were 
converted from GTF to BED12 using standard tools (gtfToGenePred and genePredToBed).
"""

import pytest
from pathlib import Path
from pyrion.io.gtf import read_gtf
from pyrion.io.bed import read_bed12_file


@pytest.mark.io
class TestGTFBEDEquivalence:
    """Test equivalence between GTF and BED12 parsing results."""
    
    @pytest.fixture
    def gtf_file_path(self):
        """Path to test GTF file."""
        return Path("test_data/gtf/sample_gencode.gtf")
    
    @pytest.fixture
    def bed_file_path(self):
        """Path to equivalent BED12 file."""
        return Path("test_data/gtf/sample_gencode.ucsc.bed")
    
    @pytest.fixture
    def gtf_transcripts(self, gtf_file_path):
        """Load transcripts from GTF file."""
        return read_gtf(str(gtf_file_path))
    
    @pytest.fixture 
    def bed_transcripts(self, bed_file_path):
        """Load transcripts from BED12 file."""
        return read_bed12_file(str(bed_file_path))
    
    def test_files_exist(self, gtf_file_path, bed_file_path):
        """Test that both test files exist."""
        assert gtf_file_path.exists(), f"GTF file not found: {gtf_file_path}"
        assert bed_file_path.exists(), f"BED12 file not found: {bed_file_path}"
        
        assert gtf_file_path.stat().st_size > 0, "GTF file is empty"
        assert bed_file_path.stat().st_size > 0, "BED12 file is empty"
    
    def test_both_parsers_return_transcripts(self, gtf_transcripts, bed_transcripts):
        """Test that both parsers successfully return transcripts."""
        assert len(gtf_transcripts) > 0, "GTF parser returned no transcripts"
        assert len(bed_transcripts) > 0, "BED12 parser returned no transcripts"
        
        # Should have reasonable number of transcripts
        assert len(gtf_transcripts) > 10, f"Expected >10 GTF transcripts, got {len(gtf_transcripts)}"
        assert len(bed_transcripts) > 10, f"Expected >10 BED12 transcripts, got {len(bed_transcripts)}"
    
    def test_transcript_id_overlap(self, gtf_transcripts, bed_transcripts):
        """Test that GTF and BED12 files contain overlapping transcript IDs."""
        gtf_ids = {t.id for t in gtf_transcripts.transcripts}
        bed_ids = {t.id for t in bed_transcripts.transcripts}
        
        # Should have some overlap (though not necessarily 100% identical)
        overlap = gtf_ids.intersection(bed_ids)
        
        assert len(overlap) > 0, "No common transcript IDs found between GTF and BED12"
        
        # Log for debugging
        overlap_ratio = len(overlap) / min(len(gtf_ids), len(bed_ids))
        
        # Expect at least some reasonable overlap
        assert overlap_ratio > 0.1, f"Low overlap ratio: {overlap_ratio:.2%}"
    
    def test_equivalent_transcript_structure(self, gtf_transcripts, bed_transcripts):
        """Test that equivalent transcripts have the same genomic structure."""
        gtf_by_id = {t.id: t for t in gtf_transcripts.transcripts}
        bed_by_id = {t.id: t for t in bed_transcripts.transcripts}
        
        # Find common transcript IDs
        common_ids = set(gtf_by_id.keys()).intersection(bed_by_id.keys())
        
        assert len(common_ids) > 0, "No common transcript IDs for structure comparison"
        
        # Test first 10 common transcripts
        test_ids = list(common_ids)[:10]
        
        for transcript_id in test_ids:
            gtf_transcript = gtf_by_id[transcript_id]
            bed_transcript = bed_by_id[transcript_id]
            
            # Test basic properties
            assert gtf_transcript.chrom == bed_transcript.chrom, \
                f"Chromosome mismatch for {transcript_id}: GTF={gtf_transcript.chrom}, BED={bed_transcript.chrom}"
            
            assert gtf_transcript.strand == bed_transcript.strand, \
                f"Strand mismatch for {transcript_id}: GTF={gtf_transcript.strand}, BED={bed_transcript.strand}"
            
            # Test exon structure (blocks)
            gtf_blocks = sorted(gtf_transcript.blocks.tolist(), key=lambda x: x[0])
            bed_blocks = sorted(bed_transcript.blocks.tolist(), key=lambda x: x[0])
            
            assert len(gtf_blocks) == len(bed_blocks), \
                f"Exon count mismatch for {transcript_id}: GTF={len(gtf_blocks)}, BED={len(bed_blocks)}"
            
            # Compare exon coordinates (allowing for small differences due to format specifics)
            for i, (gtf_block, bed_block) in enumerate(zip(gtf_blocks, bed_blocks)):
                gtf_start, gtf_end = gtf_block
                bed_start, bed_end = bed_block
                
                # Allow for coordinate system differences (0-based vs 1-based)
                start_diff = abs(gtf_start - bed_start)
                end_diff = abs(gtf_end - bed_end)
                
                assert start_diff <= 1, \
                    f"Exon {i} start mismatch for {transcript_id}: GTF={gtf_start}, BED={bed_start}"
                
                assert end_diff <= 1, \
                    f"Exon {i} end mismatch for {transcript_id}: GTF={gtf_end}, BED={bed_end}"
    
    def test_transcript_span_consistency(self, gtf_transcripts, bed_transcripts):
        """Test that transcript spans are consistent between GTF and BED12."""
        gtf_by_id = {t.id: t for t in gtf_transcripts.transcripts}
        bed_by_id = {t.id: t for t in bed_transcripts.transcripts}
        
        common_ids = set(gtf_by_id.keys()).intersection(bed_by_id.keys())
        
        # Test first 5 common transcripts
        test_ids = list(common_ids)[:5]
        
        for transcript_id in test_ids:
            gtf_transcript = gtf_by_id[transcript_id]
            bed_transcript = bed_by_id[transcript_id]
            
            # Compare transcript spans
            gtf_span = gtf_transcript.transcript_span
            bed_span = bed_transcript.transcript_span
            
            # Allow for small coordinate differences due to coordinate system (0-based vs 1-based)
            start_diff = abs(gtf_span[0] - bed_span[0])
            end_diff = abs(gtf_span[1] - bed_span[1])
            
            assert start_diff <= 1, \
                f"Transcript span start mismatch for {transcript_id}: GTF={gtf_span[0]}, BED={bed_span[0]} (diff={start_diff})"
            
            assert end_diff <= 1, \
                f"Transcript span end mismatch for {transcript_id}: GTF={gtf_span[1]}, BED={bed_span[1]} (diff={end_diff})"
    
    def test_coding_status_consistency(self, gtf_transcripts, bed_transcripts):
        """Test that coding status is consistent between GTF and BED12."""
        gtf_by_id = {t.id: t for t in gtf_transcripts.transcripts}
        bed_by_id = {t.id: t for t in bed_transcripts.transcripts}
        
        common_ids = set(gtf_by_id.keys()).intersection(bed_by_id.keys())
        
        coding_mismatches = 0
        total_compared = 0
        
        for transcript_id in list(common_ids)[:20]:  # Test first 20
            gtf_transcript = gtf_by_id[transcript_id]
            bed_transcript = bed_by_id[transcript_id]
            
            gtf_is_coding = gtf_transcript.is_coding
            bed_is_coding = bed_transcript.is_coding
            
            total_compared += 1
            
            if gtf_is_coding != bed_is_coding:
                coding_mismatches += 1
        
        # Should have very low mismatch rate for equivalent files
        mismatch_rate = coding_mismatches / total_compared if total_compared > 0 else 0
        
        assert mismatch_rate < 0.1, \
            f"High coding status mismatch rate: {mismatch_rate:.2%} ({coding_mismatches}/{total_compared})"
    
    @pytest.mark.slow
    def test_comprehensive_structure_validation(self, gtf_transcripts, bed_transcripts):
        """Comprehensive validation of transcript structures (marked slow for optional execution)."""
        gtf_by_id = {t.id: t for t in gtf_transcripts.transcripts}
        bed_by_id = {t.id: t for t in bed_transcripts.transcripts}
        
        common_ids = set(gtf_by_id.keys()).intersection(bed_by_id.keys())
        
        validation_errors = []
        
        for transcript_id in common_ids:
            gtf_transcript = gtf_by_id[transcript_id]
            bed_transcript = bed_by_id[transcript_id]
            
            # Validate exon ordering and non-overlap
            try:
                gtf_blocks = sorted(gtf_transcript.blocks.tolist())
                bed_blocks = sorted(bed_transcript.blocks.tolist())
                
                # Check for overlapping exons within each transcript
                for i in range(len(gtf_blocks) - 1):
                    if gtf_blocks[i][1] >= gtf_blocks[i+1][0]:
                        validation_errors.append(f"GTF {transcript_id}: overlapping exons {i}, {i+1}")
                
                for i in range(len(bed_blocks) - 1):
                    if bed_blocks[i][1] >= bed_blocks[i+1][0]:
                        validation_errors.append(f"BED {transcript_id}: overlapping exons {i}, {i+1}")
                
                # Check that exon structures are similar
                if len(gtf_blocks) != len(bed_blocks):
                    validation_errors.append(f"{transcript_id}: exon count differs GTF={len(gtf_blocks)} vs BED={len(bed_blocks)}")
                
            except Exception as e:
                validation_errors.append(f"{transcript_id}: validation error - {str(e)}")
        
        # Report errors but don't fail unless there are too many
        if validation_errors:
            error_rate = len(validation_errors) / len(common_ids)
            
            # Log first few errors for debugging
            for error in validation_errors[:5]:
                print(f"Validation error: {error}")
            
            if len(validation_errors) > 5:
                print(f"... and {len(validation_errors) - 5} more errors")
            
            assert error_rate < 0.1, f"High validation error rate: {error_rate:.2%}"


@pytest.mark.io
class TestGTFBEDEquivalenceMetadata:
    """Test metadata consistency between GTF and BED12 parsing."""
    
    @pytest.fixture
    def gtf_transcripts(self):
        """Load transcripts from GTF file."""
        gtf_file = Path("test_data/gtf/sample_gencode.gtf")
        return read_gtf(str(gtf_file))
    
    @pytest.fixture
    def bed_transcripts(self):
        """Load transcripts from BED12 file.""" 
        bed_file = Path("test_data/hg38.v48.comprehensive.bed")
        return read_bed12_file(str(bed_file))
    
    def test_chromosome_coverage(self, gtf_transcripts, bed_transcripts):
        """Test that chromosome coverage is similar between formats."""
        gtf_chroms = set(gtf_transcripts.get_all_chromosomes())
        bed_chroms = set(bed_transcripts.get_all_chromosomes())
        
        # Should have some chromosome overlap
        common_chroms = gtf_chroms.intersection(bed_chroms)
        assert len(common_chroms) > 0, "No common chromosomes found"
        
        # Log chromosome information for debugging
        print(f"GTF chromosomes: {sorted(gtf_chroms)}")
        print(f"BED chromosomes: {sorted(bed_chroms)}")
        print(f"Common chromosomes: {sorted(common_chroms)}")
    
    def test_strand_distribution(self, gtf_transcripts, bed_transcripts):
        """Test that strand distribution is similar between formats."""
        gtf_by_id = {t.id: t for t in gtf_transcripts.transcripts}
        bed_by_id = {t.id: t for t in bed_transcripts.transcripts}
        
        common_ids = set(gtf_by_id.keys()).intersection(bed_by_id.keys())
        
        if len(common_ids) == 0:
            pytest.skip("No common transcript IDs for strand comparison")
        
        strand_matches = 0
        for transcript_id in common_ids:
            if gtf_by_id[transcript_id].strand == bed_by_id[transcript_id].strand:
                strand_matches += 1
        
        match_rate = strand_matches / len(common_ids)
        assert match_rate > 0.95, f"Low strand match rate: {match_rate:.2%}"
    
    def test_biotype_information_preservation(self, gtf_transcripts):
        """Test that GTF parsing preserves biotype information."""
        # This specifically tests the GTF enhancement we added
        
        # Should have gene data bound
        assert gtf_transcripts._gene_data is not None, "GTF should have gene data bound"
        
        gene_data = gtf_transcripts._gene_data
        
        # Should have biotype mappings
        assert gene_data.has_biotype_mapping(), "GTF should extract biotype information"
        
        # Check that biotypes are reasonable
        biotype_count = gene_data.get_biotype_count()
        assert biotype_count > 0, "Should have extracted some biotypes"
        
        # Should cover most/all transcripts
        transcript_count = len(gtf_transcripts.transcripts)
        coverage_rate = biotype_count / transcript_count
        
        assert coverage_rate > 0.5, f"Low biotype coverage: {coverage_rate:.2%}"


"""Unit tests for pyrion.ops.chains module."""

import pytest
import numpy as np
from pathlib import Path

from pyrion.core.intervals import GenomicInterval
from pyrion.core.strand import Strand
from pyrion.core.genes import Transcript
from pyrion.core.genome_alignment import GenomeAlignment
from pyrion.io.chain import read_chain_file
from pyrion.ops.chains import (
    project_intervals_through_chain,
    _project_intervals_vectorized, 
    _project_intervals_numba,
    _project_intervals_numpy,
    project_intervals_through_genome_alignment,
    project_intervals_through_genome_alignment_to_intervals,
    get_chain_target_interval,
    get_chain_query_interval,
    get_chain_t_start,
    get_chain_t_end,
    get_chain_q_start,
    get_chain_q_end,
    split_genome_alignment,
    HAS_NUMBA
)


class TestFixtures:
    """Test data fixtures."""
    
    @pytest.fixture(scope="class")
    def test_chain_file(self):
        """Path to test chain file."""
        return Path("test_data/chains/hg38.chr9.mm39.chr4.chain")
    
    @pytest.fixture(scope="class") 
    def sample_chain_file(self):
        """Path to smaller sample chain file."""
        return Path("test_data/chains/hg38.chr9.mm39.chr4.chain")
    
    @pytest.fixture(scope="class")
    def loaded_chain(self, sample_chain_file):
        """Loaded GenomeAlignment from sample data."""
        collection = read_chain_file(sample_chain_file)
        return collection.alignments[0]  # Use first alignment
    
    @pytest.fixture
    def simple_chain_blocks(self):
        """Simple synthetic chain blocks for basic testing."""
        return np.array([
            [100, 200, 1000, 1100],  # t_start, t_end, q_start, q_end
            [300, 400, 1200, 1300],
            [500, 600, 1400, 1500],
        ], dtype=np.int64)
    
    @pytest.fixture
    def simple_genome_alignment(self, simple_chain_blocks):
        """Simple GenomeAlignment for testing."""
        return GenomeAlignment(
            chain_id=1,
            score=1000,
            t_chrom="chr1",
            t_strand=1,
            t_size=1000000,
            q_chrom="chr2", 
            q_strand=1,
            q_size=1000000,
            blocks=simple_chain_blocks
        )
    
    @pytest.fixture
    def test_intervals(self):
        """Test intervals for projection."""
        return np.array([
            [150, 180],   # Overlaps first block
            [120, 220],   # Spans first block  
            [350, 380],   # Overlaps second block
            [250, 350],   # In gap between blocks
            [80, 120],    # Before all blocks
            [700, 800],   # After all blocks
        ], dtype=np.int64)
    
    @pytest.fixture
    def sample_transcripts(self):
        """Sample transcripts for split testing."""
        return [
            Transcript(
                blocks=np.array([[1000, 1200], [1300, 1500]], dtype=np.int64),
                strand=Strand.PLUS,
                chrom="chr1",
                id="transcript1"
            ),
            Transcript(
                blocks=np.array([[2000, 2300]], dtype=np.int64),
                strand=Strand.PLUS,
                chrom="chr1", 
                id="transcript2"
            )
        ]


class TestProjectIntervalsThroughChain(TestFixtures):
    """Test interval projection functions."""
    
    def test_empty_intervals(self, simple_chain_blocks):
        """Test with empty intervals array."""
        empty_intervals = np.array([], dtype=np.int64).reshape(0, 2)
        result = project_intervals_through_chain(empty_intervals, simple_chain_blocks)
        assert result == []
    
    def test_empty_chain_blocks(self):
        """Test with empty chain blocks."""
        intervals = np.array([[100, 200]], dtype=np.int64)
        empty_blocks = np.array([], dtype=np.int64).reshape(0, 4)
        result = project_intervals_through_chain(intervals, empty_blocks)
        assert len(result) == 1
        assert np.array_equal(result[0], np.array([[0, 0]], dtype=np.int64))
    
    def test_basic_projection(self, simple_chain_blocks, test_intervals):
        """Test basic interval projection."""
        results = project_intervals_through_chain(test_intervals, simple_chain_blocks)
        assert len(results) == len(test_intervals)
        
        # First interval [150, 180] should project to [1050, 1080]
        assert np.array_equal(results[0], np.array([[1050, 1080]], dtype=np.int64))
        
        # Second interval [120, 220] should project to [1020, 1120]
        assert np.array_equal(results[1], np.array([[1020, 1120]], dtype=np.int64))
    
    def test_vectorized_dispatcher(self, simple_chain_blocks, test_intervals):
        """Test vectorized implementation dispatcher."""
        result_vectorized = _project_intervals_vectorized(test_intervals, simple_chain_blocks)
        result_numpy = _project_intervals_numpy(test_intervals, simple_chain_blocks)
        
        # Results should be identical
        assert len(result_vectorized) == len(result_numpy)
        for vec_res, np_res in zip(result_vectorized, result_numpy):
            assert np.array_equal(vec_res, np_res)
    
    @pytest.mark.skipif(not HAS_NUMBA, reason="Numba not available")
    def test_numba_implementation(self, simple_chain_blocks, test_intervals):
        """Test numba implementation if available."""
        result_numba = _project_intervals_numba(test_intervals, simple_chain_blocks)
        result_numpy = _project_intervals_numpy(test_intervals, simple_chain_blocks)
        
        # Results should be identical
        assert len(result_numba) == len(result_numpy)
        for numba_res, np_res in zip(result_numba, result_numpy):
            assert np.array_equal(numba_res, np_res)
    
    def test_clamping_to_chain_bounds(self, simple_chain_blocks):
        """Test that results are clamped to chain boundaries."""
        # Interval that would project outside chain bounds
        intervals = np.array([[50, 650]], dtype=np.int64)
        results = project_intervals_through_chain(intervals, simple_chain_blocks)
        
        # Should be clamped to chain bounds [1000, 1500]
        result = results[0][0]
        assert result[0] >= 1000  # Min q_start
        assert result[1] <= 1500  # Max q_end


class TestProjectThroughGenomeAlignment(TestFixtures):
    """Test GenomeAlignment projection functions."""
    
    def test_project_through_genome_alignment(self, simple_genome_alignment, test_intervals):
        """Test projection through GenomeAlignment object."""
        results = project_intervals_through_genome_alignment(test_intervals, simple_genome_alignment)
        assert len(results) == len(test_intervals)
        assert all(isinstance(r, np.ndarray) for r in results)
    
    def test_project_to_intervals_objects(self, simple_genome_alignment):
        """Test projection to GenomicInterval objects."""
        intervals = np.array([[150, 180], [350, 380]], dtype=np.int64)
        results = project_intervals_through_genome_alignment_to_intervals(
            intervals, simple_genome_alignment
        )
        
        assert len(results) == 2
        assert all(isinstance(r, GenomicInterval) for r in results)
        assert all(r.chrom == "chr2" for r in results)  # Target chromosome
        assert all(r.strand == Strand.PLUS for r in results)
    
    def test_project_to_intervals_custom_chrom_strand(self, simple_genome_alignment):
        """Test projection with custom chromosome and strand."""
        intervals = np.array([[150, 180]], dtype=np.int64)
        results = project_intervals_through_genome_alignment_to_intervals(
            intervals, simple_genome_alignment, 
            target_chrom="custom_chr", target_strand=Strand.MINUS
        )
        
        assert len(results) == 1
        assert results[0].chrom == "custom_chr"
        assert results[0].strand == Strand.MINUS
    
    def test_filter_invalid_projections(self, simple_genome_alignment):
        """Test that invalid projections (0,0) are filtered out."""
        # Intervals outside chain coverage
        intervals = np.array([[50, 80], [150, 180]], dtype=np.int64)
        results = project_intervals_through_genome_alignment_to_intervals(
            intervals, simple_genome_alignment
        )
        
        # Should only get valid projections
        assert len(results) == 1  # Only second interval projects validly
        assert results[0].start == 1050
        assert results[0].end == 1080


class TestChainAccessors(TestFixtures):
    """Test chain coordinate accessor functions."""
    
    def test_get_chain_target_interval(self, simple_genome_alignment):
        """Test target interval extraction."""
        target = get_chain_target_interval(simple_genome_alignment)
        
        assert isinstance(target, GenomicInterval)
        assert target.chrom == "chr1"
        assert target.start == 100  # First block start
        assert target.end == 600    # Last block end 
        assert target.strand == Strand.PLUS
        assert "chain_1_target" in target.id
    
    def test_get_chain_query_interval(self, simple_genome_alignment):
        """Test query interval extraction."""
        query = get_chain_query_interval(simple_genome_alignment)
        
        assert isinstance(query, GenomicInterval)
        assert query.chrom == "chr2"
        assert query.start == 1000  # First block q_start
        assert query.end == 1500    # Last block q_end
        assert query.strand == Strand.PLUS
        assert "chain_1_query" in query.id
    
    def test_chain_coordinate_accessors(self, simple_genome_alignment):
        """Test individual coordinate accessors."""
        assert get_chain_t_start(simple_genome_alignment) == 100
        assert get_chain_t_end(simple_genome_alignment) == 600
        assert get_chain_q_start(simple_genome_alignment) == 1000
        assert get_chain_q_end(simple_genome_alignment) == 1500
    
    def test_empty_chain_raises_error(self):
        """Test that empty chains raise ValueError."""
        empty_chain = GenomeAlignment(
            chain_id=1, score=0, t_chrom="chr1", t_strand=1, t_size=1000,
            q_chrom="chr2", q_strand=1, q_size=1000,
            blocks=np.array([], dtype=np.int64).reshape(0, 4)
        )
        
        with pytest.raises(ValueError, match="Chain has no blocks"):
            get_chain_target_interval(empty_chain)
        
        with pytest.raises(ValueError, match="Chain has no blocks"):
            get_chain_query_interval(empty_chain)
            
        with pytest.raises(ValueError, match="Chain has no blocks"):
            get_chain_t_start(empty_chain)
    
    def test_none_genome_alignment_raises_error(self):
        """Test that None GenomeAlignment raises specific error."""
        with pytest.raises(ValueError, match="provided genome_alignment object is None"):
            get_chain_target_interval(None)


class TestSplitGenomeAlignment(TestFixtures):
    """Test chain splitting functionality."""
    
    def test_short_chain_no_split(self, simple_genome_alignment, sample_transcripts):
        """Test that short chains are not split."""
        # Chain length is 500, less than 1.5 * window_size (1M)
        subchains, mapping = split_genome_alignment(
            simple_genome_alignment, sample_transcripts, window_size=1_000_000
        )
        
        assert len(subchains) == 1
        assert subchains[0].child_id == 0
        assert len(mapping[0]) == 2  # Both transcripts mapped
    
    def test_split_with_transcripts(self, sample_transcripts):
        """Test splitting with transcript-guided boundaries."""
        # Create a longer chain that needs splitting
        large_blocks = np.array([
            [100, 200, 1000, 1100],
            [500_000, 500_100, 1_500_000, 1_500_100],
            [1_500_000, 1_500_100, 2_500_000, 2_500_100],
            [2_000_000, 2_000_100, 3_000_000, 3_000_100],
        ], dtype=np.int64)
        
        large_chain = GenomeAlignment(
            chain_id=2, score=2000, t_chrom="chr1", t_strand=1, t_size=10_000_000,
            q_chrom="chr2", q_strand=1, q_size=10_000_000, blocks=large_blocks
        )
        
        subchains, mapping = split_genome_alignment(
            large_chain, sample_transcripts, window_size=800_000
        )
        
        assert len(subchains) >= 1
        assert all(sc.child_id is not None for sc in subchains)
        assert all(sc.chain_id == 2 for sc in subchains)  # Same chain_id
    
    def test_split_empty_transcripts(self):
        """Test splitting with no transcripts."""
        large_blocks = np.array([
            [100, 200, 1000, 1100],
            [1_000_000, 1_000_100, 2_000_000, 2_000_100],
            [2_000_000, 2_000_100, 3_000_000, 3_000_100],
        ], dtype=np.int64)
        
        large_chain = GenomeAlignment(
            chain_id=3, score=3000, t_chrom="chr1", t_strand=1, t_size=10_000_000,
            q_chrom="chr2", q_strand=1, q_size=10_000_000, blocks=large_blocks
        )
        
        subchains, mapping = split_genome_alignment(
            large_chain, [], window_size=500_000
        )
        
        assert len(subchains) >= 2  # Should split based on window size only
        assert all(len(mapping[i]) == 0 for i in mapping)  # No transcripts


class TestRealDataIntegration(TestFixtures):
    """Integration tests with real chain data."""
    
    @pytest.mark.io
    def test_with_real_chain_data(self, loaded_chain):
        """Test functions with real chain data."""
        assert loaded_chain.chain_id is not None
        assert len(loaded_chain.blocks) > 0
        
        # Test accessors
        t_start = get_chain_t_start(loaded_chain)
        t_end = get_chain_t_end(loaded_chain) 
        q_start = get_chain_q_start(loaded_chain)
        q_end = get_chain_q_end(loaded_chain)
        
        assert t_start >= 0
        assert t_end > t_start
        assert q_start >= 0
        assert q_end > q_start
        
        # Test interval creation
        target_interval = get_chain_target_interval(loaded_chain)
        query_interval = get_chain_query_interval(loaded_chain)
        
        assert target_interval.start == t_start
        assert target_interval.end == t_end
        assert query_interval.start == q_start
        assert query_interval.end == q_end
    
    @pytest.mark.io
    def test_projection_with_real_data(self, loaded_chain):
        """Test interval projection with real chain data."""
        # Create test intervals within chain bounds
        t_start = get_chain_t_start(loaded_chain)
        t_end = get_chain_t_end(loaded_chain)
        
        # Small intervals within chain
        test_intervals = np.array([
            [t_start + 1000, t_start + 1100],
            [t_start + 5000, t_start + 5200],
        ], dtype=np.int64)
        
        results = project_intervals_through_chain(test_intervals, loaded_chain.blocks)
        
        assert len(results) == 2
        assert all(len(r) > 0 for r in results)
        assert all(r.shape[1] == 2 for r in results)  # Each result has start,end


class TestEdgeCases(TestFixtures):
    """Test edge cases and error conditions."""
    
    def test_single_base_intervals(self, simple_chain_blocks):
        """Test projection of single-base intervals."""
        intervals = np.array([[150, 151], [350, 351]], dtype=np.int64)
        results = project_intervals_through_chain(intervals, simple_chain_blocks)
        
        assert len(results) == 2
        assert all(r[0][1] > r[0][0] for r in results)  # End > start
    
    def test_intervals_in_gaps(self, simple_chain_blocks):
        """Test intervals that fall in gaps between blocks."""
        gap_intervals = np.array([[250, 280]], dtype=np.int64)  # Between blocks 1 and 2
        results = project_intervals_through_chain(gap_intervals, simple_chain_blocks)
        
        assert len(results) == 1
        # Should get gap coordinates
        result = results[0][0]
        assert result[0] == 1100  # End of previous block
        assert result[1] == 1200  # Start of next block
    
    def test_strand_handling(self):
        """Test different strand configurations."""
        # Negative strand chain
        neg_blocks = np.array([[100, 200, 1000, 1100]], dtype=np.int64)
        neg_chain = GenomeAlignment(
            chain_id=4, score=1000, t_chrom="chr1", t_strand=-1, t_size=1000000,
            q_chrom="chr2", q_strand=-1, q_size=1000000, blocks=neg_blocks
        )
        
        target = get_chain_target_interval(neg_chain)
        query = get_chain_query_interval(neg_chain)
        
        assert target.strand == Strand.MINUS
        assert query.strand == Strand.MINUS
    
    def test_large_intervals(self, simple_chain_blocks):
        """Test projection of very large intervals."""
        large_intervals = np.array([[0, 1000000]], dtype=np.int64)
        results = project_intervals_through_chain(large_intervals, simple_chain_blocks)
        
        assert len(results) == 1
        # Should be clamped to chain bounds
        result = results[0][0]
        assert result[0] >= 1000  # Min bound
        assert result[1] <= 1500  # Max bound





"""Test core properties, utilities, and algorithmic edge cases.

This module focuses on testing:
- Core transcript and gene properties that are fundamental to genomics analysis
- Utility functions for constants, amino acid analysis, and data type handling  
- Algorithm threshold behaviors and performance-critical code paths
- Edge cases and error handling in core data structures
"""

import pytest
import numpy as np
from pyrion.core.strand import Strand
from pyrion.core.genes import Transcript
from pyrion.core.amino_acid_sequences import AminoAcidSequence
from pyrion.core_types import ExonType
from pyrion.ops.interval_ops import merge_intervals, intersect_intervals
from pyrion.ops.interval_collection_ops import merge_close_intervals
from pyrion.core.intervals import GenomicInterval


class TestGenomicsConstants:
    """Test genomics constants and enumerations."""
    
    def test_exon_prefixes(self):
        """Test EXON_PREFIXES constants."""
        from pyrion.constants import EXON_PREFIXES
        
        assert EXON_PREFIXES[ExonType.ALL] == "exon"
        assert EXON_PREFIXES[ExonType.CDS] == "cds_exon"
        assert EXON_PREFIXES[ExonType.UTR5] == "utr5_exon"
        assert EXON_PREFIXES[ExonType.UTR3] == "utr3_exon"
        assert len(EXON_PREFIXES) == 4
    
    def test_exon_numbering_constants(self):
        """Test exon numbering constants."""
        from pyrion.constants import EXON_NUMBER_START, EXON_ID_SEPARATOR
        
        assert EXON_NUMBER_START == 1
        assert EXON_ID_SEPARATOR == ":"


class TestAminoAcidAnalysis:
    """Test amino acid sequence analysis functions."""
    
    def test_count_amino_acids_in_sequence(self):
        """Test amino acid counting function."""
        from pyrion.core.amino_acid_auxiliary import count_amino_acids_in_sequence
        
        # Create a simple amino acid sequence
        aa_seq = AminoAcidSequence.from_string("ACDEFGHIKLMNPQRSTVWY")
        
        counts = count_amino_acids_in_sequence(aa_seq)
        
        # Should have 20 different amino acids, each with count 1
        assert isinstance(counts, dict)
        assert len(counts) == 20
        assert all(count == 1 for count in counts.values())
        assert "A" in counts
        assert "C" in counts
        assert "Y" in counts
    
    def test_count_amino_acids_with_duplicates(self):
        """Test amino acid counting with repeated amino acids."""
        from pyrion.core.amino_acid_auxiliary import count_amino_acids_in_sequence
        
        aa_seq = AminoAcidSequence.from_string("AAACCCGGG")
        counts = count_amino_acids_in_sequence(aa_seq)
        
        assert counts["A"] == 3
        assert counts["C"] == 3
        assert counts["G"] == 3
        assert len(counts) == 3


class TestTranscriptCoreProperties:
    """Test fundamental Transcript properties for genomic analysis."""
    
    @pytest.fixture
    def sample_transcript(self):
        """Create a sample transcript for testing."""
        return Transcript(
            blocks=np.array([[1000, 1200], [1500, 1700]], dtype=np.int32),
            strand=Strand.PLUS,
            chrom="chr1",
            id="TEST001",
            cds_start=1050,
            cds_end=1650,
            biotype="protein_coding"
        )
    
    def test_transcript_span(self, sample_transcript):
        """Test transcript_span property."""
        span = sample_transcript.transcript_span
        
        assert isinstance(span, np.ndarray)
        assert span.shape == (2,)
        assert span[0] == 1000  # Start of first block
        assert span[1] == 1700  # End of last block
    
    def test_transcript_interval(self, sample_transcript):
        """Test transcript_interval property."""
        interval = sample_transcript.transcript_interval
        
        assert isinstance(interval, GenomicInterval)
        assert interval.chrom == "chr1"
        assert interval.start == 1000
        assert interval.end == 1700
        assert interval.strand == Strand.PLUS
        assert interval.id == "TEST001"
    
    def test_transcript_cds_interval(self, sample_transcript):
        """Test transcript_cds_interval property."""
        cds_interval = sample_transcript.transcript_cds_interval
        
        assert isinstance(cds_interval, GenomicInterval)
        assert cds_interval.chrom == "chr1"
        assert cds_interval.start == 1050
        assert cds_interval.end == 1650
        assert cds_interval.strand == Strand.PLUS
    
    def test_non_coding_transcript_cds_interval(self):
        """Test CDS interval for non-coding transcript."""
        non_coding = Transcript(
            blocks=np.array([[1000, 1200]], dtype=np.int32),
            strand=Strand.PLUS,
            chrom="chr1",
            id="NC001"
        )
        
        cds_interval = non_coding.transcript_cds_interval
        assert cds_interval is None


class TestPerformanceCriticalPaths:
    """Test performance-critical algorithm paths and large dataset handling."""
    
    def test_merge_intervals_numba_threshold(self):
        """Test merge_intervals with large dataset to trigger numba."""
        # Create dataset larger than numba threshold (usually ~20k)
        large_intervals = np.random.randint(0, 100000, (25000, 2))
        large_intervals = np.sort(large_intervals, axis=1)  # Ensure start < end
        large_intervals = large_intervals[np.argsort(large_intervals[:, 0])]  # Sort by start
        
        # This should trigger the numba implementation
        result = merge_intervals(large_intervals, use_numba=True)
        
        assert isinstance(result, np.ndarray)
        assert result.shape[1] == 2
        assert len(result) <= len(large_intervals)  # Merged result should be smaller or equal
    
    def test_intersect_intervals_numba_threshold(self):
        """Test intersect_intervals with large dataset to trigger numba."""
        # Create two large datasets
        intervals1 = np.random.randint(0, 50000, (15000, 2))
        intervals1 = np.sort(intervals1, axis=1)
        intervals2 = np.random.randint(25000, 75000, (15000, 2))
        intervals2 = np.sort(intervals2, axis=1)
        
        # This should trigger the numba implementation
        result = intersect_intervals(intervals1, intervals2, use_numba=True)
        
        assert isinstance(result, np.ndarray)
        assert result.shape[1] == 2
    
    def test_merge_close_intervals_numba_path(self):
        """Test merge_close_intervals with conditions that trigger numba helpers."""
        from pyrion.core.intervals import GenomicIntervalsCollection
        
        # Create large collection with small gaps that will merge
        intervals = []
        for i in range(1000):
            start = i * 60  # 60bp spacing
            end = start + 50  # 50bp intervals, so 10bp gaps
            intervals.append(GenomicInterval("chr1", start, end, Strand.PLUS, f"interval_{i}"))
        
        collection = GenomicIntervalsCollection.from_intervals(intervals)
        
        # This should exercise the numba helper functions and merge intervals (10bp gaps < 15bp threshold)
        result = merge_close_intervals(collection, max_gap=15)
        
        assert isinstance(result, GenomicIntervalsCollection)
        assert len(result) < len(collection)  # Should merge some intervals


class TestGenomicsUtilities:
    """Test core genomics utility functions and data type operations."""
    
    def test_strand_utilities(self):
        """Test strand utility functions."""
        # Test from_int method
        plus_strand = Strand.from_int(1)
        minus_strand = Strand.from_int(-1)
        
        assert plus_strand == Strand.PLUS
        assert minus_strand == Strand.MINUS
        
        # Test to_char method
        assert Strand.PLUS.to_char() == "+"
        assert Strand.MINUS.to_char() == "-"
    
    def test_encoding_edge_cases(self):
        """Test encoding utility edge cases."""
        from pyrion.utils.encoding import encode_nucleotides
        
        # Test with mixed case
        result = encode_nucleotides("AtCg")
        assert isinstance(result, np.ndarray)
        assert len(result) == 4
        
        # Test with gaps
        result_with_gaps = encode_nucleotides("AT-GC")
        assert len(result_with_gaps) == 5


class TestGenomicsDataTypes:
    """Test genomics-specific data types and enumerations."""
    
    def test_core_types_enum_usage(self):
        """Test ExonType enum in realistic scenarios."""
        from pyrion.core_types import ExonType
        
        # Test all enum values are accessible
        assert ExonType.ALL is not None
        assert ExonType.CDS is not None
        assert ExonType.UTR5 is not None
        assert ExonType.UTR3 is not None
        
        # Test enum can be used in collections
        exon_types = [ExonType.ALL, ExonType.CDS]
        assert len(exon_types) == 2
        assert ExonType.ALL in exon_types


class TestEdgeCasesAndErrorHandling:
    """Test edge cases and error handling in core genomics operations."""
    
    def test_empty_transcript_blocks_error(self):
        """Test transcript with empty blocks."""
        with pytest.raises(IndexError):
            # This should fail when trying to access blocks[0] in transcript_span
            empty_transcript = Transcript(
                blocks=np.empty((0, 2), dtype=np.int32),
                strand=Strand.PLUS,
                chrom="chr1",
                id="EMPTY001"
            )
            _ = empty_transcript.transcript_span
    
    def test_invalid_interval_operations(self):
        """Test interval operations with invalid inputs."""
        # Test with malformed intervals
        invalid_intervals = np.array([[100, 50]], dtype=np.int32)  # end < start
        
        # Should still work but produce empty result
        result = merge_intervals(invalid_intervals)
        assert isinstance(result, np.ndarray)


"""Unit tests for pyrion.ops.interval_ops module."""

import pytest
import numpy as np
from typing import List

from pyrion.ops.interval_ops import (
    merge_intervals,
    _merge_intervals_numba,
    _merge_intervals_numpy,
    intersect_intervals,
    _intersect_intervals_numba,
    subtract_intervals,
    _subtract_intervals_numba,
    intervals_union
)


class TestFixtures:
    """Test data fixtures for interval operations."""
    
    @pytest.fixture
    def overlapping_intervals(self):
        """Overlapping intervals for merge testing."""
        return np.array([
            [10, 30],
            [25, 50],
            [100, 150],
            [140, 200],
            [300, 400]
        ], dtype=np.int32)
    
    @pytest.fixture
    def non_overlapping_intervals(self):
        """Non-overlapping intervals."""
        return np.array([
            [10, 20],
            [30, 40],
            [50, 60],
            [100, 110]
        ], dtype=np.int32)
    
    @pytest.fixture
    def adjacent_intervals(self):
        """Adjacent (touching) intervals."""
        return np.array([
            [10, 20],
            [20, 30],
            [30, 40],
            [50, 60]
        ], dtype=np.int32)
    
    @pytest.fixture
    def complex_intervals(self):
        """Complex overlapping pattern for thorough testing."""
        return np.array([
            [1, 10],
            [5, 15],
            [12, 25],
            [20, 30],
            [35, 45],
            [40, 50],
            [60, 70]
        ], dtype=np.int32)
    
    @pytest.fixture
    def intersecting_sets(self):
        """Two sets of intervals for intersection testing."""
        set1 = np.array([
            [10, 50],
            [100, 150],
            [200, 300]
        ], dtype=np.int32)
        
        set2 = np.array([
            [30, 70],
            [120, 180],
            [250, 350],
            [400, 500]  # No intersection with set1
        ], dtype=np.int32)
        
        return set1, set2


class TestMergeIntervals(TestFixtures):
    """Test interval merging functionality."""
    
    def test_merge_overlapping_intervals(self, overlapping_intervals):
        """Test merging of overlapping intervals."""
        result = merge_intervals(overlapping_intervals)
        
        expected = np.array([
            [10, 50],    # Merged [10,30] and [25,50]
            [100, 200],  # Merged [100,150] and [140,200]
            [300, 400]   # No merge needed
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_merge_non_overlapping_intervals(self, non_overlapping_intervals):
        """Test merging of non-overlapping intervals."""
        result = merge_intervals(non_overlapping_intervals)
        
        # Should return sorted intervals unchanged
        expected = np.array([
            [10, 20],
            [30, 40],
            [50, 60],
            [100, 110]
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_merge_adjacent_intervals(self, adjacent_intervals):
        """Test merging of adjacent (touching) intervals."""
        result = merge_intervals(adjacent_intervals)
        
        expected = np.array([
            [10, 40],  # Merged [10,20], [20,30], [30,40]
            [50, 60]   # No merge needed
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_merge_complex_pattern(self, complex_intervals):
        """Test merging of complex overlapping pattern."""
        result = merge_intervals(complex_intervals)
        
        expected = np.array([
            [1, 30],   # Merged [1,10], [5,15], [12,25], [20,30]
            [35, 50],  # Merged [35,45], [40,50]
            [60, 70]   # No merge needed
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_merge_single_interval(self):
        """Test merging single interval."""
        single = np.array([[100, 200]], dtype=np.int32)
        result = merge_intervals(single)
        
        assert np.array_equal(result, single)
    
    def test_merge_empty_array(self):
        """Test merging empty array."""
        empty = np.array([], dtype=np.int32).reshape(0, 2)
        result = merge_intervals(empty)
        
        assert result.shape == (0, 2)
        assert result.dtype == np.int32
    
    def test_merge_unsorted_intervals(self):
        """Test merging unsorted intervals."""
        unsorted = np.array([
            [100, 150],
            [10, 30],
            [25, 50],
            [200, 300]
        ], dtype=np.int32)
        
        result = merge_intervals(unsorted)
        
        expected = np.array([
            [10, 50],     # Merged and sorted
            [100, 150],
            [200, 300]
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_merge_numba_vs_numpy(self, complex_intervals):
        """Test that numba and numpy implementations give same results."""
        result_numba = _merge_intervals_numba(complex_intervals)
        result_numpy = _merge_intervals_numpy(complex_intervals)
        
        assert np.array_equal(result_numba, result_numpy)
    
    def test_merge_algorithm_selection(self, overlapping_intervals):
        """Test automatic algorithm selection."""
        # Force numba
        result_numba = merge_intervals(overlapping_intervals, use_numba=True)
        # Force numpy
        result_numpy = merge_intervals(overlapping_intervals, use_numba=False)
        # Auto-select
        result_auto = merge_intervals(overlapping_intervals, use_numba=None)
        
        assert np.array_equal(result_numba, result_numpy)
        assert np.array_equal(result_numpy, result_auto)


class TestIntersectIntervals(TestFixtures):
    """Test interval intersection functionality."""
    
    def test_basic_intersections(self, intersecting_sets):
        """Test basic interval intersections."""
        set1, set2 = intersecting_sets
        result = intersect_intervals(set1, set2)
        
        expected = np.array([
            [30, 50],   # [10,50] ∩ [30,70]
            [120, 150], # [100,150] ∩ [120,180]
            [250, 300]  # [200,300] ∩ [250,350]
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_no_intersections(self):
        """Test intervals with no intersections."""
        set1 = np.array([[10, 20], [100, 110]], dtype=np.int32)
        set2 = np.array([[30, 40], [200, 210]], dtype=np.int32)
        
        result = intersect_intervals(set1, set2)
        
        assert result.shape == (0, 2)
        assert result.dtype == np.int32
    
    def test_complete_overlap(self):
        """Test intervals with complete overlap."""
        set1 = np.array([[10, 100]], dtype=np.int32)
        set2 = np.array([[20, 30], [40, 60]], dtype=np.int32)
        
        result = intersect_intervals(set1, set2)
        
        expected = np.array([
            [20, 30],
            [40, 60]
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_partial_overlaps(self):
        """Test various partial overlap scenarios."""
        set1 = np.array([[10, 50], [100, 200]], dtype=np.int32)
        set2 = np.array([[30, 80], [150, 250]], dtype=np.int32)
        
        result = intersect_intervals(set1, set2)
        
        expected = np.array([
            [30, 50],   # Partial overlap
            [150, 200]  # Partial overlap
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_intersect_empty_arrays(self):
        """Test intersection with empty arrays."""
        intervals = np.array([[100, 200]], dtype=np.int32)
        empty = np.array([], dtype=np.int32).reshape(0, 2)
        
        # Empty second set
        result = intersect_intervals(intervals, empty)
        assert result.shape == (0, 2)
        
        # Empty first set
        result = intersect_intervals(empty, intervals)
        assert result.shape == (0, 2)
        
        # Both empty
        result = intersect_intervals(empty, empty)
        assert result.shape == (0, 2)
    
    def test_intersect_numba_vs_numpy(self, intersecting_sets):
        """Test that numba and numpy implementations give same results."""
        set1, set2 = intersecting_sets
        
        result_numba = intersect_intervals(set1, set2, use_numba=True)
        result_numpy = intersect_intervals(set1, set2, use_numba=False)
        
        # Sort results for comparison (order might differ)
        result_numba_sorted = result_numba[np.lexsort((result_numba[:, 1], result_numba[:, 0]))]
        result_numpy_sorted = result_numpy[np.lexsort((result_numpy[:, 1], result_numpy[:, 0]))]
        
        assert np.array_equal(result_numba_sorted, result_numpy_sorted)
    
    def test_intersect_identical_sets(self):
        """Test intersection of identical interval sets."""
        intervals = np.array([[10, 20], [30, 40], [50, 60]], dtype=np.int32)
        result = intersect_intervals(intervals, intervals)
        
        # Should return the original intervals
        assert np.array_equal(result, intervals)


class TestSubtractIntervals(TestFixtures):
    """Test interval subtraction functionality."""
    
    def test_basic_subtraction(self):
        """Test basic interval subtraction."""
        base = np.array([[10, 100]], dtype=np.int32)
        subtract = np.array([[30, 70]], dtype=np.int32)
        
        result = subtract_intervals(base, subtract)
        
        expected = np.array([
            [10, 30],   # Left part
            [70, 100]   # Right part
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_subtraction_no_overlap(self):
        """Test subtraction with no overlapping intervals."""
        base = np.array([[10, 20], [100, 110]], dtype=np.int32)
        subtract = np.array([[30, 40], [200, 210]], dtype=np.int32)
        
        result = subtract_intervals(base, subtract)
        
        # Should return original intervals unchanged
        assert np.array_equal(result, base)
    
    def test_complete_subtraction(self):
        """Test complete removal of intervals."""
        base = np.array([[10, 50], [100, 150]], dtype=np.int32)
        subtract = np.array([[5, 60], [90, 160]], dtype=np.int32)  # Completely covers base
        
        result = subtract_intervals(base, subtract)
        
        # Should return empty result
        assert result.shape == (0, 2)
    
    def test_partial_subtraction(self):
        """Test partial subtraction scenarios."""
        base = np.array([
            [10, 100],
            [200, 300],
            [400, 500]
        ], dtype=np.int32)
        
        subtract = np.array([
            [30, 70],    # Splits first interval
            [250, 350],  # Partial overlap with second
            [600, 700]   # No overlap with any
        ], dtype=np.int32)
        
        result = subtract_intervals(base, subtract)
        
        expected = np.array([
            [10, 30],    # Left part of first
            [70, 100],   # Right part of first
            [200, 250],  # Left part of second
            [400, 500]   # Third unchanged
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_multiple_subtractions_single_interval(self):
        """Test multiple subtractions from a single interval."""
        base = np.array([[10, 100]], dtype=np.int32)
        subtract = np.array([
            [20, 30],
            [40, 50],
            [70, 80]
        ], dtype=np.int32)
        
        result = subtract_intervals(base, subtract)
        
        expected = np.array([
            [10, 20],
            [30, 40],
            [50, 70],
            [80, 100]
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_subtraction_edge_cases(self):
        """Test edge cases in subtraction."""
        base = np.array([[10, 50]], dtype=np.int32)
        
        # Subtract from left edge
        subtract_left = np.array([[10, 20]], dtype=np.int32)
        result = subtract_intervals(base, subtract_left)
        expected = np.array([[20, 50]], dtype=np.int32)
        assert np.array_equal(result, expected)
        
        # Subtract from right edge
        subtract_right = np.array([[40, 50]], dtype=np.int32)
        result = subtract_intervals(base, subtract_right)
        expected = np.array([[10, 40]], dtype=np.int32)
        assert np.array_equal(result, expected)
    
    def test_subtraction_empty_arrays(self):
        """Test subtraction with empty arrays."""
        intervals = np.array([[100, 200]], dtype=np.int32)
        empty = np.array([], dtype=np.int32).reshape(0, 2)
        
        # Empty subtract array - should return original
        result = subtract_intervals(intervals, empty)
        assert np.array_equal(result, intervals)
        
        # Empty base array - should return empty
        result = subtract_intervals(empty, intervals)
        assert result.shape == (0, 2)
    
    def test_subtraction_numba_vs_numpy(self):
        """Test that numba and numpy implementations give same results."""
        base = np.array([[10, 100], [200, 300]], dtype=np.int32)
        subtract = np.array([[30, 70], [250, 350]], dtype=np.int32)
        
        result_numba = subtract_intervals(base, subtract, use_numba=True)
        result_numpy = subtract_intervals(base, subtract, use_numba=False)
        
        # Sort results for comparison
        result_numba_sorted = result_numba[np.lexsort((result_numba[:, 1], result_numba[:, 0]))]
        result_numpy_sorted = result_numpy[np.lexsort((result_numpy[:, 1], result_numpy[:, 0]))]
        
        assert np.array_equal(result_numba_sorted, result_numpy_sorted)


class TestIntervalsUnion(TestFixtures):
    """Test interval union functionality."""
    
    def test_basic_union(self):
        """Test basic union of multiple interval sets."""
        set1 = np.array([[10, 30], [100, 150]], dtype=np.int32)
        set2 = np.array([[25, 50], [200, 250]], dtype=np.int32)
        set3 = np.array([[40, 80], [140, 180]], dtype=np.int32)
        
        result = intervals_union([set1, set2, set3])
        
        expected = np.array([
            [10, 80],    # Merged [10,30], [25,50], [40,80]
            [100, 180],  # Merged [100,150], [140,180]
            [200, 250]   # Standalone
        ], dtype=np.int32)
        
        assert np.array_equal(result, expected)
    
    def test_union_single_set(self):
        """Test union with single interval set."""
        intervals = np.array([[10, 30], [50, 80]], dtype=np.int32)
        result = intervals_union([intervals])
        
        assert np.array_equal(result, intervals)
    
    def test_union_empty_sets(self):
        """Test union with empty sets."""
        set1 = np.array([[10, 20]], dtype=np.int32)
        empty = np.array([], dtype=np.int32).reshape(0, 2)
        
        # Mix of empty and non-empty
        result = intervals_union([empty, set1, empty])
        assert np.array_equal(result, set1)
        
        # All empty
        result = intervals_union([empty, empty])
        assert result.shape == (0, 2)
        
        # Empty list
        result = intervals_union([])
        assert result.shape == (0, 2)
    
    def test_union_overlapping_sets(self):
        """Test union of heavily overlapping sets."""
        set1 = np.array([[10, 50]], dtype=np.int32)
        set2 = np.array([[30, 70]], dtype=np.int32)
        set3 = np.array([[60, 100]], dtype=np.int32)
        
        result = intervals_union([set1, set2, set3])
        
        expected = np.array([[10, 100]], dtype=np.int32)
        assert np.array_equal(result, expected)
    
    def test_union_many_sets(self):
        """Test union of many small sets."""
        # Create many small non-overlapping intervals
        sets = [
            np.array([[i*10, i*10 + 5]], dtype=np.int32) 
            for i in range(10)
        ]
        
        result = intervals_union(sets)
        
        # Should have 10 separate intervals
        assert len(result) == 10
        assert result[0][0] == 0 and result[0][1] == 5
        assert result[-1][0] == 90 and result[-1][1] == 95


class TestValidationAndErrorHandling(TestFixtures):
    """Test input validation and error handling."""
    
    def test_invalid_array_shapes(self):
        """Test error handling for invalid array shapes."""
        invalid_shape = np.array([[100, 200, 300]], dtype=np.int32)  # Wrong shape
        valid_shape = np.array([[100, 200]], dtype=np.int32)
        
        with pytest.raises(ValueError, match="intervals must have shape"):
            merge_intervals(invalid_shape)
        
        with pytest.raises(ValueError, match="intervals must have shape"):
            intersect_intervals(invalid_shape, valid_shape)
        
        with pytest.raises(ValueError, match="intervals must have shape"):
            subtract_intervals(valid_shape, invalid_shape)
    
    def test_data_type_handling(self):
        """Test handling of different data types."""
        # Float input should be converted to int32
        float_intervals = np.array([[10.5, 20.7], [30.1, 40.9]], dtype=np.float64)
        
        result = merge_intervals(float_intervals)
        assert result.dtype == np.int32
        assert np.array_equal(result, np.array([[10, 20], [30, 40]], dtype=np.int32))
    
    def test_large_datasets(self):
        """Test performance with larger datasets."""
        # Create large dataset
        n = 10000
        intervals = np.random.randint(0, 1000000, size=(n, 2))
        intervals = np.sort(intervals, axis=1)  # Ensure start < end
        
        # Test that operations complete without error
        merged = merge_intervals(intervals)
        assert merged.dtype == np.int32
        assert len(merged) <= n  # Should be merged down
        
        # Test intersection of large sets
        subset1 = intervals[:n//2]
        subset2 = intervals[n//2:]
        intersected = intersect_intervals(subset1, subset2)
        assert intersected.dtype == np.int32


class TestPerformanceBenchmarks(TestFixtures):
    """Performance comparison tests."""
    
    def test_merge_performance_comparison(self):
        """Compare numba vs numpy merge performance."""
        # Create moderately large dataset
        n = 5000
        intervals = np.random.randint(0, 100000, size=(n, 2))
        intervals = np.sort(intervals, axis=1)
        
        # Both implementations should work
        result_numba = _merge_intervals_numba(intervals)
        result_numpy = _merge_intervals_numpy(intervals)
        
        # Results should be equivalent (after sorting)
        result_numba_sorted = result_numba[np.lexsort((result_numba[:, 1], result_numba[:, 0]))]
        result_numpy_sorted = result_numpy[np.lexsort((result_numpy[:, 1], result_numpy[:, 0]))]
        
        assert np.array_equal(result_numba_sorted, result_numpy_sorted)
    
    def test_intersect_performance_comparison(self):
        """Compare numba vs numpy intersect performance."""
        # Create test datasets
        n1, n2 = 1000, 800
        set1 = np.random.randint(0, 50000, size=(n1, 2))
        set1 = np.sort(set1, axis=1)
        set2 = np.random.randint(0, 50000, size=(n2, 2))
        set2 = np.sort(set2, axis=1)
        
        # Both implementations should work
        result_numba = intersect_intervals(set1, set2, use_numba=True)
        result_numpy = intersect_intervals(set1, set2, use_numba=False)
        
        # Results should be equivalent (after sorting)
        if len(result_numba) > 0 and len(result_numpy) > 0:
            result_numba_sorted = result_numba[np.lexsort((result_numba[:, 1], result_numba[:, 0]))]
            result_numpy_sorted = result_numpy[np.lexsort((result_numpy[:, 1], result_numpy[:, 0]))]
            assert np.array_equal(result_numba_sorted, result_numpy_sorted)





"""Unit tests for pyrion.ops.intervals module."""

import pytest
import numpy as np
from typing import List

from pyrion.core.intervals import GenomicInterval
from pyrion.core.strand import Strand
from pyrion.core.genes import Transcript
from pyrion.core.genome_alignment import GenomeAlignment
from pyrion.ops.intervals import (
    find_intersections,
    compute_intersections_core,
    compute_overlap_size,
    intervals_to_array,
    array_to_intervals,
    chains_to_arrays,
    transcripts_to_arrays,
    projected_intervals_to_genomic_intervals
)


class TestFixtures:
    """Test data fixtures for interval operations."""
    
    @pytest.fixture
    def sample_intervals_array(self):
        """Sample interval array for testing."""
        return np.array([
            [100, 200],
            [150, 250],
            [300, 400],
            [350, 450],
            [500, 600]
        ], dtype=np.int32)
    
    @pytest.fixture
    def sample_intervals_objects(self):
        """Sample GenomicInterval objects."""
        return [
            GenomicInterval(chrom="chr1", start=100, end=200, strand=Strand.PLUS, id="int1"),
            GenomicInterval(chrom="chr1", start=300, end=400, strand=Strand.PLUS, id="int2"),
            GenomicInterval(chrom="chr1", start=500, end=600, strand=Strand.MINUS, id="int3")
        ]
    
    @pytest.fixture
    def sample_transcripts(self):
        """Sample transcripts for testing."""
        return [
            Transcript(
                blocks=np.array([[100, 200], [300, 400]], dtype=np.int64),
                strand=Strand.PLUS,
                chrom="chr1",
                id="transcript1"
            ),
            Transcript(
                blocks=np.array([[1000, 1200]], dtype=np.int64),
                strand=Strand.MINUS,
                chrom="chr2", 
                id="transcript2"
            )
        ]
    
    @pytest.fixture
    def sample_chains(self):
        """Sample genome alignments (chains) for testing."""
        return [
            GenomeAlignment(
                chain_id=1, score=1000, t_chrom="chr1", t_strand=1, t_size=1000000,
                q_chrom="chr2", q_strand=1, q_size=1000000,
                blocks=np.array([[100, 200, 1000, 1100], [300, 400, 1200, 1300]], dtype=np.int64)
            ),
            GenomeAlignment(
                chain_id=2, score=2000, t_chrom="chr3", t_strand=-1, t_size=2000000,
                q_chrom="chr4", q_strand=-1, q_size=2000000,
                blocks=np.array([[500, 700, 2000, 2200]], dtype=np.int64)
            )
        ]
    
    @pytest.fixture
    def overlapping_arrays(self):
        """Two arrays with overlapping intervals."""
        arr1 = np.array([[10, 50], [100, 150], [200, 300]], dtype=np.int32)
        arr2 = np.array([[30, 80], [120, 180], [250, 350]], dtype=np.int32)
        return arr1, arr2


class TestFindIntersections(TestFixtures):
    """Test interval intersection functionality."""
    
    def test_basic_intersections(self, overlapping_arrays):
        """Test basic intersection finding."""
        arr1, arr2 = overlapping_arrays
        intersections = find_intersections(arr1, arr2)
        
        # Should have intersections for all three intervals in arr1
        assert len(intersections) == 3
        assert 0 in intersections  # First interval
        assert 1 in intersections  # Second interval
        assert 2 in intersections  # Third interval
        
        # Check overlap values
        assert intersections[0][0][1] == 20  # [10,50] ∩ [30,80] = 20bp
        assert intersections[1][0][1] == 30  # [100,150] ∩ [120,180] = 30bp
        assert intersections[2][0][1] == 50  # [200,300] ∩ [250,350] = 50bp
    
    def test_intersections_with_ids(self, overlapping_arrays):
        """Test intersections with custom IDs."""
        arr1, arr2 = overlapping_arrays
        ids1 = ["interval_A", "interval_B", "interval_C"]
        ids2 = ["target_X", "target_Y", "target_Z"]
        
        intersections = find_intersections(arr1, arr2, ids1, ids2)
        
        assert "interval_A" in intersections
        assert "interval_B" in intersections
        assert "interval_C" in intersections
        
        # Check that target IDs are preserved
        assert intersections["interval_A"][0][0] == "target_X"
        assert intersections["interval_B"][0][0] == "target_Y"
        assert intersections["interval_C"][0][0] == "target_Z"
    
    def test_empty_arrays(self):
        """Test intersection with empty arrays."""
        arr1 = np.array([[100, 200]], dtype=np.int32)
        empty_arr = np.array([], dtype=np.int32).reshape(0, 2)
        
        # Empty second array
        result = find_intersections(arr1, empty_arr)
        assert result == {}
        
        # Empty first array
        result = find_intersections(empty_arr, arr1)
        assert result == {}
        
        # Both empty
        result = find_intersections(empty_arr, empty_arr)
        assert result == {}
    
    def test_no_intersections(self):
        """Test arrays with no overlapping intervals."""
        arr1 = np.array([[10, 20], [100, 110]], dtype=np.int32)
        arr2 = np.array([[30, 40], [200, 210]], dtype=np.int32)
        
        result = find_intersections(arr1, arr2)
        assert result == {}
    
    def test_invalid_input_shapes(self):
        """Test error handling for invalid input shapes."""
        invalid_arr = np.array([[100, 200, 300]], dtype=np.int32)  # Wrong shape
        valid_arr = np.array([[100, 200]], dtype=np.int32)
        
        with pytest.raises(ValueError, match="Arrays must have shape"):
            find_intersections(invalid_arr, valid_arr)
    
    def test_mismatched_ids(self):
        """Test error handling for mismatched ID lengths."""
        arr1 = np.array([[100, 200], [300, 400]], dtype=np.int32)
        arr2 = np.array([[150, 250]], dtype=np.int32)
        wrong_ids = ["id1"]  # Should be length 2
        
        with pytest.raises(ValueError, match="ids1 length must match"):
            find_intersections(arr1, arr2, ids1=wrong_ids)


class TestComputeIntersectionsCore(TestFixtures):
    """Test the numba-optimized core intersection function."""
    
    def test_core_intersection_function(self):
        """Test the core numba intersection implementation."""
        sorted1 = np.array([[10, 50], [100, 150]], dtype=np.int32)
        sorted2 = np.array([[30, 80], [120, 180]], dtype=np.int32)
        idx1 = np.array([0, 1], dtype=np.int32)
        idx2 = np.array([0, 1], dtype=np.int32)
        
        i_arr, j_arr, ov_arr = compute_intersections_core(sorted1, sorted2, idx1, idx2)
        
        assert len(i_arr) == 2  # Two intersections
        assert len(j_arr) == 2
        assert len(ov_arr) == 2
        
        # Check overlaps
        assert ov_arr[0] == 20  # [10,50] ∩ [30,80] = 20
        assert ov_arr[1] == 30  # [100,150] ∩ [120,180] = 30
    
    def test_core_no_intersections(self):
        """Test core function with no overlapping intervals."""
        sorted1 = np.array([[10, 20]], dtype=np.int32)
        sorted2 = np.array([[30, 40]], dtype=np.int32)
        idx1 = np.array([0], dtype=np.int32)
        idx2 = np.array([0], dtype=np.int32)
        
        i_arr, j_arr, ov_arr = compute_intersections_core(sorted1, sorted2, idx1, idx2)
        
        assert len(i_arr) == 0
        assert len(j_arr) == 0
        assert len(ov_arr) == 0


class TestComputeOverlapSize(TestFixtures):
    """Test overlap size computation."""
    
    def test_overlap_calculation(self):
        """Test basic overlap size calculation."""
        # Overlapping intervals
        assert compute_overlap_size(10, 50, 30, 80) == 20  # [10,50] ∩ [30,80]
        assert compute_overlap_size(100, 200, 150, 250) == 50  # [100,200] ∩ [150,250]
        
        # Non-overlapping intervals
        assert compute_overlap_size(10, 20, 30, 40) == 0
        assert compute_overlap_size(100, 110, 50, 90) == 0
        
        # Adjacent intervals (no overlap)
        assert compute_overlap_size(10, 20, 20, 30) == 0
        
        # Contained intervals
        assert compute_overlap_size(10, 100, 30, 50) == 20  # [30,50] contained in [10,100]
        assert compute_overlap_size(30, 50, 10, 100) == 20  # Same result
    
    def test_identical_intervals(self):
        """Test overlap of identical intervals."""
        assert compute_overlap_size(100, 200, 100, 200) == 100
    
    def test_single_base_overlap(self):
        """Test minimal overlap scenarios."""
        assert compute_overlap_size(10, 21, 20, 30) == 1  # Single base overlap


class TestArrayConversions(TestFixtures):
    """Test array conversion functions."""
    
    def test_intervals_to_array(self, sample_intervals_objects):
        """Test conversion from GenomicInterval objects to array."""
        array = intervals_to_array(sample_intervals_objects)
        
        expected = np.array([
            [100, 200],
            [300, 400], 
            [500, 600]
        ], dtype=np.int32)
        
        assert np.array_equal(array, expected)
    
    def test_intervals_to_array_empty(self):
        """Test conversion of empty interval list."""
        array = intervals_to_array([])
        assert array.shape == (0, 2)
        assert array.dtype == np.int32
    
    def test_array_to_intervals(self, sample_intervals_array):
        """Test conversion from array to GenomicInterval objects."""
        intervals = array_to_intervals(sample_intervals_array, "chr5")
        
        assert len(intervals) == 5
        assert all(iv.chrom == "chr5" for iv in intervals)
        assert all(iv.strand == Strand.UNKNOWN for iv in intervals)
        
        # Check coordinates
        assert intervals[0].start == 100 and intervals[0].end == 200
        assert intervals[2].start == 300 and intervals[2].end == 400
    
    def test_array_to_intervals_empty(self):
        """Test conversion of empty array."""
        empty_array = np.array([], dtype=np.int32).reshape(0, 2)
        intervals = array_to_intervals(empty_array, "chr1")
        assert intervals == []


class TestChainArrayConversion(TestFixtures):
    """Test chain-specific array conversion functions."""
    
    def test_chains_to_arrays_target(self, sample_chains):
        """Test conversion of chains to target coordinate arrays."""
        spans, ids = chains_to_arrays(sample_chains, for_q=False)
        
        assert spans.shape == (2, 2)
        assert len(ids) == 2
        
        # Check target spans
        assert spans[0][0] == 100 and spans[0][1] == 400  # Chain 1 target span
        assert spans[1][0] == 500 and spans[1][1] == 700  # Chain 2 target span
        
        # Check chain IDs
        assert ids[0] == 1
        assert ids[1] == 2
    
    def test_chains_to_arrays_query(self, sample_chains):
        """Test conversion of chains to query coordinate arrays."""
        spans, ids = chains_to_arrays(sample_chains, for_q=True)
        
        assert spans.shape == (2, 2)
        
        # Check query spans (note: handles strand automatically)
        assert spans[0][0] == 1000 and spans[0][1] == 1300  # Chain 1 query span
        # Chain 2 is negative strand, so coordinates should be handled differently
        assert len(spans[1]) == 2
    
    def test_chains_to_arrays_empty(self):
        """Test conversion of empty chain list."""
        spans, ids = chains_to_arrays([])
        
        assert spans.shape == (0, 2)
        assert len(ids) == 0
        assert spans.dtype == np.int32
        assert ids.dtype == np.int32


class TestTranscriptArrayConversion(TestFixtures):
    """Test transcript-specific array conversion functions."""
    
    def test_transcripts_to_arrays(self, sample_transcripts):
        """Test conversion of transcripts to arrays."""
        spans, ids = transcripts_to_arrays(sample_transcripts)
        
        assert spans.shape == (2, 2)
        assert len(ids) == 2
        
        # Check transcript spans
        assert spans[0][0] == 100 and spans[0][1] == 400  # Transcript 1 span
        assert spans[1][0] == 1000 and spans[1][1] == 1200  # Transcript 2 span
        
        # Check IDs
        assert ids[0] == "transcript1"
        assert ids[1] == "transcript2"
    
    def test_transcripts_to_arrays_empty(self):
        """Test conversion of empty transcript list."""
        spans, ids = transcripts_to_arrays([])
        
        assert spans.shape == (0, 2)
        assert len(ids) == 0
        assert spans.dtype == np.int32
        assert ids.dtype == object


class TestProjectedIntervalsConversion(TestFixtures):
    """Test conversion of projected interval arrays to GenomicInterval objects."""
    
    def test_projected_intervals_conversion(self):
        """Test conversion of projected arrays to GenomicInterval objects."""
        # Simulate output from projection function
        projected_arrays = [
            np.array([[1000, 1100], [1200, 1300]], dtype=np.int64),  # Multiple results
            np.array([[2000, 2100]], dtype=np.int64),                # Single result
            np.array([[0, 0]], dtype=np.int64),                      # Invalid (unmappable)
            np.array([], dtype=np.int64).reshape(0, 2)               # Empty
        ]
        
        result = projected_intervals_to_genomic_intervals(
            projected_arrays, "chr4", Strand.PLUS, ids=["input1", "input2", "input3", "input4"]
        )
        
        assert len(result) == 4
        
        # First input - multiple results
        assert len(result[0]) == 2
        assert result[0][0].chrom == "chr4"
        assert result[0][0].start == 1000 and result[0][0].end == 1100
        assert result[0][0].strand == Strand.PLUS
        assert result[0][0].id == "input1_0"  # Multi-result ID
        assert result[0][1].id == "input1_1"
        
        # Second input - single result
        assert len(result[1]) == 1
        assert result[1][0].id == "input2"  # Single result ID
        
        # Third input - invalid (0,0) filtered out
        assert len(result[2]) == 0
        
        # Fourth input - empty
        assert len(result[3]) == 0
    
    def test_projected_intervals_no_ids(self):
        """Test conversion without providing IDs."""
        projected_arrays = [
            np.array([[1000, 1100]], dtype=np.int64),
            np.array([[2000, 2100]], dtype=np.int64)
        ]
        
        result = projected_intervals_to_genomic_intervals(
            projected_arrays, "chr5", Strand.MINUS
        )
        
        assert len(result) == 2
        assert result[0][0].id is None
        assert result[1][0].id is None
        assert all(iv[0].strand == Strand.MINUS for iv in result)
    
    def test_projected_intervals_edge_cases(self):
        """Test edge cases in projected interval conversion."""
        # All invalid results
        projected_arrays = [
            np.array([[0, 0]], dtype=np.int64),
            np.array([[0, 0], [0, 0]], dtype=np.int64)
        ]
        
        result = projected_intervals_to_genomic_intervals(
            projected_arrays, "chr1", Strand.UNKNOWN
        )
        
        assert len(result) == 2
        assert len(result[0]) == 0  # All filtered out
        assert len(result[1]) == 0  # All filtered out


class TestRealDataIntegration(TestFixtures):
    """Integration tests with more realistic data."""
    
    def test_large_dataset_intersections(self):
        """Test intersections with larger datasets."""
        # Create larger test dataset
        n1, n2 = 1000, 800
        arr1 = np.random.randint(0, 100000, size=(n1, 2))
        arr1 = np.sort(arr1, axis=1)  # Ensure start < end
        
        arr2 = np.random.randint(0, 100000, size=(n2, 2))
        arr2 = np.sort(arr2, axis=1)
        
        # Test that function runs without error
        intersections = find_intersections(arr1, arr2)
        
        # Basic sanity checks
        assert isinstance(intersections, dict)
        assert all(isinstance(k, (int, np.integer)) for k in intersections.keys())
    
    def test_conversion_round_trip(self, sample_intervals_objects):
        """Test round-trip conversion: objects -> array -> objects."""
        # Forward conversion
        array = intervals_to_array(sample_intervals_objects)
        
        # Backward conversion
        converted_back = array_to_intervals(array, "chr1")
        
        # Check that coordinates match (ignoring IDs/strands which aren't preserved)
        for orig, converted in zip(sample_intervals_objects, converted_back):
            assert orig.start == converted.start
            assert orig.end == converted.end
            assert orig.chrom == converted.chrom


class TestPerformanceAndEdgeCases(TestFixtures):
    """Test performance characteristics and edge cases."""
    
    def test_single_interval_arrays(self):
        """Test functions with single-interval arrays."""
        single_interval = np.array([[100, 200]], dtype=np.int32)
        
        # Test intersections
        intersections = find_intersections(single_interval, single_interval)
        assert 0 in intersections
        assert intersections[0][0][1] == 100  # Full overlap
        
        # Test conversions
        intervals = array_to_intervals(single_interval, "chr1")
        assert len(intervals) == 1
        
        back_to_array = intervals_to_array(intervals)
        assert np.array_equal(back_to_array, single_interval)
    
    def test_zero_length_intervals(self):
        """Test handling of zero-length intervals."""
        # Note: GenomicInterval validation should prevent these, but test robustness
        zero_length = np.array([[100, 100]], dtype=np.int32)
        normal = np.array([[50, 150]], dtype=np.int32)
        
        # Should not crash, though behavior may vary
        result = find_intersections(zero_length, normal)
        assert isinstance(result, dict)
    
    def test_very_large_coordinates(self):
        """Test with very large genomic coordinates."""
        large_coords = np.array([
            [1000000000, 1000001000],  # 1Gb region
            [2000000000, 2000002000]   # 2Gb region
        ], dtype=np.int32)
        
        # Test basic operations
        overlap = compute_overlap_size(1000000500, 1000001500, 1000000000, 1000001000)
        assert overlap == 500
        
        intervals = array_to_intervals(large_coords, "chr1")
        assert len(intervals) == 2
        assert intervals[0].start == 1000000000





"""Canonizer functions for selecting canonical transcripts from transcript lists."""

from typing import List, Optional
import numpy as np


def longest_isoform_canonizer(transcripts: List, **kwargs) -> Optional[str]:
    """Default canonizer that selects the transcript with the longest total exonic length."""
    if not transcripts:
        return None
    
    if len(transcripts) == 1:
        return transcripts[0].id
    
    max_length = 0
    canonical_transcript_id = None
    
    for transcript in transcripts:
        exon_lengths = transcript.blocks[:, 1] - transcript.blocks[:, 0]
        total_length = int(np.sum(exon_lengths))
        
        if total_length > max_length:
            max_length = total_length
            canonical_transcript_id = transcript.id
    
    return canonical_transcript_id


def longest_cds_canonizer(transcripts: List, **kwargs) -> Optional[str]:
    if not transcripts:
        return None
    
    coding_transcripts = [t for t in transcripts if t.is_coding]
    
    if not coding_transcripts:
        return longest_isoform_canonizer(transcripts, **kwargs)
    
    if len(coding_transcripts) == 1:
        return coding_transcripts[0].id

    max_cds_length = 0
    canonical_transcript_id = None
    
    for transcript in coding_transcripts:
        cds_blocks = transcript.cds_blocks
        if len(cds_blocks) > 0:
            cds_lengths = cds_blocks[:, 1] - cds_blocks[:, 0]
            total_cds_length = int(np.sum(cds_lengths))
            
            if total_cds_length > max_cds_length:
                max_cds_length = total_cds_length
                canonical_transcript_id = transcript.id
    
    return canonical_transcript_id


def longest_transcript_span_canonizer(transcripts: List, **kwargs) -> Optional[str]:
    if not transcripts:
        return None
    
    if len(transcripts) == 1:
        return transcripts[0].id
    
    max_span = 0
    canonical_transcript_id = None
    
    for transcript in transcripts:
        span = transcript.blocks[-1, 1] - transcript.blocks[0, 0]
        
        if span > max_span:
            max_span = span
            canonical_transcript_id = transcript.id
    
    return canonical_transcript_id


def first_transcript_canonizer(transcripts: List, **kwargs) -> Optional[str]:
    if not transcripts:
        return None
    return transcripts[0].id


def most_exons_canonizer(transcripts: List, **kwargs) -> Optional[str]:
    if not transcripts:
        return None
    
    if len(transcripts) == 1:
        return transcripts[0].id
    
    # Find transcript with most exons
    max_exons = 0
    canonical_transcript_id = None
    
    for transcript in transcripts:
        num_exons = len(transcript.blocks)
        
        if num_exons > max_exons:
            max_exons = num_exons
            canonical_transcript_id = transcript.id
    
    return canonical_transcript_id

DEFAULT_CANONIZER = longest_isoform_canonizer



from typing import List

import numpy as np


def _vstack_safe(arrays: List[np.ndarray]) -> np.ndarray:
    non_empty = [a for a in arrays if a.size > 0]
    return np.vstack(non_empty) if non_empty else np.empty((0, 2), dtype=int)



"""Gene and transcript representations."""

from dataclasses import dataclass
from functools import cached_property
from typing import Optional, List, Dict, Set, Tuple, Union, Callable
import numpy as np
from weakref import WeakKeyDictionary
from pathlib import Path

from .genes_auxiliary import build_annotated_regions
from .strand import Strand
from .intervals import GenomicInterval, AnnotatedIntervalSet
from . import genes_auxiliary

_region_cache = WeakKeyDictionary()


@dataclass(frozen=True)
class Transcript:
    blocks: np.ndarray  # (N, 2) array: start, end
    strand: Strand
    chrom: str
    id: str
    cds_start: Optional[int] = None
    cds_end: Optional[int] = None
    biotype: Optional[str] = None

    @property
    def is_coding(self) -> bool:
        return self.cds_start is not None and self.cds_end is not None

    def __hash__(self) -> int:
        return hash(self.id)

    @cached_property
    def transcript_span(self) -> np.ndarray:
        """Get genomic span of transcript (min start, max end) regardless of block order."""
        return np.array([self.blocks[:, 0].min(), self.blocks[:, 1].max()], dtype=np.int32)

    @cached_property
    def transcript_interval(self) -> GenomicInterval:
        return genes_auxiliary.get_transcript_interval(self)

    @cached_property
    def transcript_cds_interval(self) -> GenomicInterval:
        return genes_auxiliary.get_transcript_cds_interval(self)

    @cached_property
    def cds_blocks(self) -> np.ndarray:
        return genes_auxiliary.get_cds_blocks(self)

    @cached_property
    def utr5_blocks(self) -> np.ndarray:
        return genes_auxiliary.get_utr5_blocks(self)

    @cached_property
    def utr3_blocks(self) -> np.ndarray:
        return genes_auxiliary.get_utr3_blocks(self)
    
    @cached_property
    def utr_blocks(self) -> np.ndarray:
        """Get all UTR intervals (merged 5' and 3')."""
        utr5 = self.utr5_blocks
        utr3 = self.utr3_blocks
        
        if utr5.size == 0 and utr3.size == 0:
            return np.empty((0, 2), dtype=np.uint32)
        elif utr5.size == 0:
            return utr3
        elif utr3.size == 0:
            return utr5
        else:
            return np.vstack([utr5, utr3])

    def __str__(self) -> str:
        from ..ops.transcript_serialization import transcript_to_bed12_string
        return transcript_to_bed12_string(self)
    
    def __repr__(self) -> str:
        coding_info = f", coding={self.is_coding}" if self.is_coding else ", non-coding"
        cds_info = f", CDS={self.cds_start}-{self.cds_end}" if self.is_coding else ""
        
        strand_str = Strand(self.strand).to_char()
            
        return f"Transcript(id='{self.id}', {self.chrom}:{self.blocks[0,0]}-{self.blocks[-1,1]}:{strand_str}, {len(self.blocks)} exons{coding_info}{cds_info})"

    def compute_flanks(
            self,
            flank_size: int,
            chrom_sizes: Dict[str, int]
    ) -> Tuple[Optional[GenomicInterval], Optional[GenomicInterval]]:
        return genes_auxiliary.compute_flanks(self, flank_size, chrom_sizes)

    def get_annotated_regions(self, chrom_sizes: dict, flank_size: int = 5000) -> AnnotatedIntervalSet:
        cache = _region_cache.setdefault(self, {})
        key = (flank_size, id(chrom_sizes))
        if key not in cache:
            cache[key] = build_annotated_regions(self, chrom_sizes, flank_size)
        return cache[key]
    
    def contains_interval(self, interval: GenomicInterval) -> bool:
        if self.chrom != interval.chrom:
            return False
        
        block_starts = self.blocks[:, 0]
        block_ends = self.blocks[:, 1]
        
        no_overlap = (interval.end <= block_starts) | (block_ends <= interval.start)
        has_overlap = ~no_overlap
        
        return np.any(has_overlap)
    
    def splice_junctions(self):
        """Generator yielding splice junction coordinates (donor, acceptor) for transcript."""
        if len(self.blocks) <= 1:
            # Single exon transcript has no splice junctions
            return
        
        sorted_blocks = self.blocks[np.argsort(self.blocks[:, 0])]
        for i in range(len(sorted_blocks) - 1):
            current_end = sorted_blocks[i, 1]      # Donor site (end of current exon)
            next_start = sorted_blocks[i + 1, 0]   # Acceptor site (start of next exon)
            
            if self.strand == Strand.MINUS:
                yield (next_start, current_end)  # (acceptor, donor)
            else:
                yield (current_end, next_start)  # (donor, acceptor)
                
    def get_introns(self, use_numba: bool = True) -> np.ndarray:
        if len(self.blocks) <= 1:
            return np.empty((0, 2), dtype=np.int32)
        
        intron_starts = self.blocks[:-1, 1]
        intron_ends = self.blocks[1:, 0]
        valid_mask = intron_starts < intron_ends
        
        if not np.any(valid_mask):
            return np.empty((0, 2), dtype=np.int32)
            
        return np.column_stack((intron_starts[valid_mask], intron_ends[valid_mask]))


class Gene:
    """Gene containing multiple transcripts with computed genomic bounds."""
    def __init__(self, gene_id: str, transcripts: List[Transcript], gene_name: Optional[str] = None):
        self.gene_id = gene_id
        self.gene_name = gene_name
        self._transcripts = list(transcripts)
        self._canonical_transcript_id: Optional[str] = None
        
        if not self._transcripts:
            raise ValueError(f"Gene {gene_id} must have at least one transcript")
        
        # Validate all transcripts are on same chromosome and strand
        first_transcript = self._transcripts[0]
        self._chrom = first_transcript.chrom
        self._strand = first_transcript.strand
        
        for transcript in self._transcripts[1:]:
            if transcript.chrom != self._chrom:
                raise ValueError(f"All transcripts must be on same chromosome. Found {transcript.chrom} and {self._chrom}")
            if transcript.strand != self._strand:
                raise ValueError(f"All transcripts must have same strand. Found {transcript.strand} and {self._strand}")
    
    @property
    def transcripts(self) -> List[Transcript]:
        return self._transcripts.copy()
    
    @property
    def transcript_ids(self) -> Set[str]:
        return {t.id for t in self._transcripts}
    
    @property
    def chrom(self) -> str:
        return self._chrom

    @property
    def strand(self) -> Strand:
        return self._strand
    
    @cached_property
    def start(self) -> int:
        return int(min(t.transcript_span[0] for t in self._transcripts)) if self._transcripts else 0
    
    @cached_property
    def end(self) -> int:
        return int(max(t.transcript_span[1] for t in self._transcripts)) if self._transcripts else 0

    @cached_property
    def bounds(self) -> GenomicInterval:
        return GenomicInterval(self.chrom, self.start, self.end, self.strand)

    @property
    def length(self) -> int:
        return self.end - self.start
    
    def get_transcript(self, transcript_id: str) -> Optional[Transcript]:
        for transcript in self._transcripts:
            if transcript.id == transcript_id:
                return transcript
        return None
    
    def has_transcript(self, transcript_id: str) -> bool:
        return transcript_id in self.transcript_ids
    
    @property
    def is_coding(self) -> bool:
        """Check if gene has any coding transcripts."""
        return any(t.is_coding for t in self._transcripts)
    
    def apply_canonizer(self, canonizer_func: Optional[Callable] = None, **kwargs) -> None:
        """Set the canonical transcript using a canonizer function.
        
        Args:
            canonizer_func: Function that takes transcripts list and returns canonical transcript ID.
                          If None, uses the default longest_isoform_canonizer.
            **kwargs: Additional arguments passed to the canonizer function.
        """
        if canonizer_func is None:
            from .canonizer import DEFAULT_CANONIZER
            canonizer_func = DEFAULT_CANONIZER
        
        canonical_id = canonizer_func(self._transcripts, **kwargs)
        
        if canonical_id is not None and canonical_id not in self.transcript_ids:
            raise ValueError(f"Canonical transcript ID '{canonical_id}' not found in gene {self.gene_id}")
        
        self._canonical_transcript_id = canonical_id


    def set_canonical_transcript(self, transcript_id: str) -> None:
        if transcript_id not in self.transcript_ids:
            raise ValueError(f"Transcript ID '{transcript_id}' not found in gene {self.gene_id}")
        
        self._canonical_transcript_id = transcript_id
    
    @property
    def canonical_transcript_id(self) -> Optional[str]:
        return self._canonical_transcript_id
    
    @property
    def canonical_transcript(self) -> Optional[Transcript]:
        if self._canonical_transcript_id is None:
            return None
        return self.get_transcript(self._canonical_transcript_id)
    
    @property
    def has_canonical_transcript(self) -> bool:
        return self._canonical_transcript_id is not None
    
    def clear_canonical_transcript(self) -> None:
        self._canonical_transcript_id = None
    
    def __len__(self) -> int:
        return len(self._transcripts)
    
    def __repr__(self) -> str:
        name_info = f", name={self.gene_name}" if self.gene_name else ""
        canonical_info = f", canonical={self._canonical_transcript_id}" if self.has_canonical_transcript else ""
        return f"Gene(id={self.gene_id}{name_info}, {self.chrom}:{self.start}-{self.end}:{self._strand}, {len(self._transcripts)} transcripts{canonical_info})"


class TranscriptsCollection:
    """Container for many transcripts."""

    def __init__(
        self,
        transcripts: Optional[List[Transcript]] = None,
        source_file: Optional[str] = None
    ):
        self.transcripts: List[Transcript] = transcripts or []
        self.source_file: Optional[str] = source_file

        self._id_index: Optional[Dict[str, int]] = None  # transcript_id → index
        self._chrom_index: Optional[Dict[str, List[int]]] = None  # chrom → indices
        self._gene_data: Optional['GeneData'] = None
        self._genes_cache: Optional[Dict[str, Gene]] = None
        self._applied_biotypes: bool = False
        self._applied_gene_names: bool = False

    def __len__(self):
        return len(self.transcripts)

    def __getitem__(self, idx: int) -> Transcript:
        return self.transcripts[idx]

    def get_by_id(self, transcript_id: str) -> Optional[Transcript]:
        if self._id_index is None:
            self._build_id_index()
        idx = self._id_index.get(transcript_id)
        return self.transcripts[idx] if idx is not None else None

    def get_by_chrom(self, chrom: str) -> List[Transcript]:
        if self._chrom_index is None:
            self._build_chrom_index()
        indices = self._chrom_index.get(chrom, [])
        return [self.transcripts[i] for i in indices]

    def get_all_chromosomes(self) -> List[str]:
        if self._chrom_index is None:
            self._build_chrom_index()
        return list(self._chrom_index.keys())

    def get_transcript_ids_by_chrom(self, chrom: str) -> List[str]:
        if self._chrom_index is None:
            self._build_chrom_index()
        indices = self._chrom_index.get(chrom, [])
        return [self.transcripts[i].id for i in indices]
    
    def get_transcripts_in_interval(self, interval: GenomicInterval, include_partial: bool = True) -> 'TranscriptsCollection':
        from .genes_auxiliary import filter_transcripts_in_interval
        return filter_transcripts_in_interval(self, interval, include_partial)

    def _build_id_index(self):
        self._id_index = {t.id: i for i, t in enumerate(self.transcripts)}

    def _build_chrom_index(self):
        from collections import defaultdict
        chrom_index = defaultdict(list)
        for i, t in enumerate(self.transcripts):
            chrom_index[t.chrom].append(i)
        self._chrom_index = dict(chrom_index)

    def bind_gene_data(self, gene_data: 'GeneData') -> None:
        self._gene_data = gene_data
        self._genes_cache = None  # Clear cache when binding new data
        self._applied_biotypes = False
        self._applied_gene_names = False
        
        # Apply transcript biotypes if available
        if gene_data.has_biotype_mapping():
            self._apply_transcript_biotypes()
        
        # Apply gene names if we have gene-transcript mapping and gene names
        if gene_data.has_gene_transcript_mapping() and gene_data.has_gene_name_mapping():
            self._applied_gene_names = True  # Will be applied when building genes cache

    def _apply_transcript_biotypes(self) -> None:
        if self._gene_data is None or not self._gene_data.has_biotype_mapping():
            return
        
        updated_transcripts = []
        for transcript in self.transcripts:
            biotype = self._gene_data.get_transcript_biotype(transcript.id)
            if biotype is not None and transcript.biotype is None:
                updated_transcript = Transcript(
                    blocks=transcript.blocks,
                    strand=transcript.strand,
                    chrom=transcript.chrom,
                    id=transcript.id,
                    cds_start=transcript.cds_start,
                    cds_end=transcript.cds_end,
                    biotype=biotype
                )
                updated_transcripts.append(updated_transcript)
            else:
                updated_transcripts.append(transcript)
        
        self.transcripts = updated_transcripts
        self._applied_biotypes = True
        self._id_index = None
        self._chrom_index = None

    def _build_genes_cache(self):
        if not self.has_gene_mapping:
            raise ValueError("No gene data with gene-transcript mapping bound. Call bind_gene_data() first.")
        
        if self._id_index is None:
            self._build_id_index()
        
        genes_dict = {}
        
        for gene_id in self._gene_data.gene_ids:
            # Get transcripts for this gene that exist in our collection
            transcript_ids = self._gene_data.get_transcripts(gene_id)
            gene_transcripts = []
            
            for transcript_id in transcript_ids:
                transcript = self.get_by_id(transcript_id)
                if transcript is not None:
                    gene_transcripts.append(transcript)
            
            if gene_transcripts:
                gene_name = None
                if self._applied_gene_names and self._gene_data.has_gene_name_mapping():
                    gene_name = self._gene_data.get_gene_name(gene_id)
                
                genes_dict[gene_id] = Gene(gene_id, gene_transcripts, gene_name=gene_name)
        
        self._genes_cache = genes_dict

    def get_gene_by_id(self, gene_id: str) -> Optional[Gene]:
        if not self.has_gene_mapping:
            raise ValueError("No gene data with gene-transcript mapping bound. Call bind_gene_data() first.")
        
        if self._genes_cache is None:
            self._build_genes_cache()
        
        return self._genes_cache.get(gene_id)

    def get_gene_by_transcript_id(self, transcript_id: str) -> Optional[Gene]:
        if not self.has_gene_mapping:
            raise ValueError("No gene data with gene-transcript mapping bound. Call bind_gene_data() first.")
        
        gene_id = self._gene_data.get_gene(transcript_id)
        if gene_id is None:
            return None
        
        return self.get_gene_by_id(gene_id)

    def get_by_gene_name(self, gene_name: str) -> List[Gene]:
        """Get Gene objects by gene name. Multiple genes can have the same name."""
        if not self.has_gene_mapping:
            raise ValueError("No gene data with gene-transcript mapping bound. Call bind_gene_data() first.")
        
        if not self._gene_data.has_gene_name_mapping():
            raise ValueError("No gene name mapping available. Gene names were not provided in the gene data.")
        
        gene_ids = self._gene_data.get_genes_by_name(gene_name)
        
        genes = []
        for gene_id in gene_ids:
            gene = self.get_gene_by_id(gene_id)
            if gene is not None:
                genes.append(gene)
        
        return genes

    @property
    def genes(self) -> List[Gene]:
        if not self.has_gene_mapping:
            raise ValueError("No gene data with gene-transcript mapping bound. Call bind_gene_data() first.")
        
        if self._genes_cache is None:
            self._build_genes_cache()
        
        return list(self._genes_cache.values())

    @property
    def gene_ids(self) -> Set[str]:
        if not self.has_gene_mapping:
            raise ValueError("No gene data with gene-transcript mapping bound. Call bind_gene_data() first.")
        
        if self._genes_cache is None:
            self._build_genes_cache()
        
        return set(self._genes_cache.keys())

    @property
    def has_gene_mapping(self) -> bool:
        return self._gene_data is not None and self._gene_data.has_gene_transcript_mapping()
    
    @property
    def available_data_mappings(self) -> List[str]:
        if self._gene_data is None:
            return []
        
        mappings = []
        if self._gene_data.has_gene_transcript_mapping():
            mappings.append("gene-transcript")
        if self._gene_data.has_biotype_mapping():
            mappings.append("transcript-biotype")
        if self._gene_data.has_gene_name_mapping():
            mappings.append("gene-name")
        return mappings
    
    @property 
    def applied_biotypes(self) -> bool:
        return self._applied_biotypes
    
    @property
    def applied_gene_names(self) -> bool:
        return self._applied_gene_names

    def canonize_transcripts(self, canonizer_func: Optional[Callable] = None, **kwargs) -> None:
        from .genes_auxiliary import set_canonical_transcripts_for_collection
        set_canonical_transcripts_for_collection(self, canonizer_func, **kwargs)

    def get_canonical_transcripts(self) -> 'TranscriptsCollection':
        from .genes_auxiliary import get_canonical_transcripts_only_from_collection
        return get_canonical_transcripts_only_from_collection(self)

    def apply_gene_canonical_mapping(self, gene_to_canonical: Dict[str, str]) -> None:
        """gene_to_canonical: Dictionary mapping gene IDs to canonical transcript IDs"""
        if not self.has_gene_mapping:
            raise ValueError("No gene data with gene-transcript mapping bound. Call bind_gene_data() first.")
        
        # Ensure genes cache is built
        if self._genes_cache is None:
            self._build_genes_cache()
        
        # Apply mapping to each gene
        for gene_id, canonical_transcript_id in gene_to_canonical.items():
            gene = self._genes_cache.get(gene_id)
            if gene is not None:
                gene.set_canonical_transcript(canonical_transcript_id)

    def get_genes_without_canonical_transcript(self) -> List[Gene]:
        if not self.has_gene_mapping:
            raise ValueError("No gene data with gene-transcript mapping bound. Call bind_gene_data() first.")
        
        if self._genes_cache is None:
            self._build_genes_cache()
        
        return [gene for gene in self._genes_cache.values() if not gene.has_canonical_transcript]

    def summary(self) -> str:
        base_summary = f"{len(self.transcripts):,} transcripts"
        if self._gene_data is not None and self._gene_data.has_gene_transcript_mapping():
            if self._genes_cache is None:
                self._build_genes_cache()
            gene_count = len(self._genes_cache)
            base_summary += f", {gene_count:,} genes"
        
        source_str = f" from {self.source_file}" if self.source_file else ""
        return base_summary + source_str

    def __str__(self) -> str:
        from ..ops.transcript_serialization import transcripts_collection_summary_string
        return transcripts_collection_summary_string(self)
    
    def save_biodata(
        self, 
        tsv_path: Union[str, Path],
        include_gene_transcript: bool = True,
        include_transcript_biotype: bool = True,
        include_gene_name: bool = True,
        separator: str = '\t'
    ) -> None:
        if self._gene_data is None:
            raise ValueError("No GeneData bound to this collection. Call bind_gene_data() first or use read_gtf() to load data with gene information.")
        
        from ..io.gene_data import write_gene_data_tsv
        write_gene_data_tsv(
            self._gene_data,
            tsv_path,
            include_gene_transcript=include_gene_transcript,
            include_transcript_biotype=include_transcript_biotype,
            include_gene_name=include_gene_name,
            separator=separator
        )

    def __repr__(self) -> str:
        return f"<TranscriptsCollection: {self.summary()}>"

    def to_bed12_string(self) -> str:
        from ..ops.transcript_serialization import transcripts_collection_to_bed12_string
        return transcripts_collection_to_bed12_string(self)

    def save_to_bed12(self, file_path: Union[str, Path]) -> None:
        from ..ops.transcript_serialization import save_transcripts_collection_to_bed12
        save_transcripts_collection_to_bed12(self, file_path)

    def save_to_json(self, file_path: Union[str, Path]) -> None:
        from ..ops.transcript_serialization import save_transcripts_collection_to_json
        save_transcripts_collection_to_json(self, file_path)

    @classmethod
    def from_json(cls, file_path: Union[str, Path]) -> 'TranscriptsCollection':
        from ..ops.transcript_serialization import load_transcripts_collection_from_json
        return load_transcripts_collection_from_json(file_path)




"""High-performance GTF format I/O support using C extension."""

import gzip
import re
from typing import Union, List, Iterator, Optional, TextIO
from pathlib import Path

from ..core.genes import TranscriptsCollection
from ..core.gene_data import GeneData
from .._gtfparser import parse_gtf_chunk


class GTFChunkReader:
    def __init__(self, file_path: Union[str, Path], chunk_size_mb: int = 512):
        self.file_path = Path(file_path)
        self.chunk_size = chunk_size_mb * 1024 * 1024  # Convert MB to bytes
        self.is_gzipped = self.file_path.suffix.lower() == '.gz'
        
    def _open_file(self) -> TextIO:
        if self.is_gzipped:
            return gzip.open(self.file_path, 'rt', encoding='utf-8')
        else:
            return open(self.file_path, 'r', encoding='utf-8')
    
    def read_gene_chunks(self) -> Iterator[List[str]]:
        with self._open_file() as f:
            current_gene_id = None
            buffer = []
            
            while True:
                chunk_text = f.read(self.chunk_size)
                if not chunk_text:
                    # End of file - yield remaining buffer
                    if buffer:
                        yield buffer
                    break
                
                lines = chunk_text.split('\n')
                if chunk_text.endswith('\n'):
                    complete_lines = lines
                    incomplete_line = ''
                else:
                    complete_lines = lines[:-1]
                    incomplete_line = lines[-1]
                
                for line in complete_lines:
                    line = line.strip()
                    if not line or line.startswith('#'):
                        continue
                    
                    gene_id = self._extract_gene_id_fast(line)
                    if not gene_id:
                        continue
                    
                    if current_gene_id is None:
                        current_gene_id = gene_id
                    
                    if gene_id == current_gene_id:
                        # Same gene - add to buffer
                        buffer.append(line)
                    else:
                        # New gene - yield current buffer and start new one
                        if buffer:
                            yield buffer
                        buffer = [line]
                        current_gene_id = gene_id
                
                # Handle incomplete line for next iteration
                if incomplete_line:
                    # Read until we get the complete line
                    rest_of_line = f.readline()
                    complete_line = incomplete_line + rest_of_line
                    
                    line = complete_line.strip()
                    if line and not line.startswith('#'):
                        gene_id = self._extract_gene_id_fast(line)
                        if gene_id:
                            if gene_id == current_gene_id:
                                buffer.append(line)
                            else:
                                if buffer:
                                    yield buffer
                                buffer = [line]
                                current_gene_id = gene_id
    
    @staticmethod
    def _extract_gene_id_fast(line: str) -> Optional[str]:
        tab_count = 0
        attr_start = 0
        
        for i, char in enumerate(line):
            if char == '\t':
                tab_count += 1
                if tab_count == 8:  # After 8th tab is the attributes column
                    attr_start = i + 1
                    break
        
        if tab_count < 8:
            return None
        
        # Look for gene_id pattern in attributes
        attrs = line[attr_start:]
        match = re.search(r'gene_id\s+"([^"]+)"', attrs)
        return match.group(1) if match else None


def read_gtf(file_path: Union[str, Path], 
             chunk_size_mb: int = 512) -> TranscriptsCollection:
    file_path = Path(file_path)
    
    if not file_path.exists():
        raise FileNotFoundError(f"GTF file not found: {file_path}")
    
    reader = GTFChunkReader(file_path, chunk_size_mb)
    all_transcripts = []
    all_gene_mapping = {}
    processed_genes = 0
    
    all_transcript_biotypes = {}
    all_gene_names = {}
    
    for chunk_lines in reader.read_gene_chunks():
        parse_result = parse_gtf_chunk(chunk_lines)
        transcripts, gene_mapping, transcript_biotypes, gene_names = parse_result
        
        all_transcripts.extend(transcripts)
        all_gene_mapping.update(gene_mapping)
        all_transcript_biotypes.update(transcript_biotypes)
        all_gene_names.update(gene_names)
        processed_genes += 1

    collection = TranscriptsCollection(transcripts=all_transcripts, source_file=str(file_path))
    gene_data = GeneData(source_file=str(file_path))
    
    for transcript_id, gene_id in all_gene_mapping.items():
        gene_data.add_gene_transcript_mapping(gene_id, transcript_id)
    
    for transcript_id, biotype in all_transcript_biotypes.items():
        gene_data.add_transcript_biotype(transcript_id, biotype)
    
    for gene_id, gene_name in all_gene_names.items():
        gene_data.add_gene_name(gene_id, gene_name)
    
    collection.bind_gene_data(gene_data)
    
    return collection 


"""Gene data storage with multiple mappings."""

from typing import Dict, Set, Optional
from collections import defaultdict


class GeneData:
    """Gene data container with optional mappings for gene-transcript relationships, biotypes, and names."""
    
    def __init__(self, source_file: Optional[str] = None):
        self.source_file: Optional[str] = source_file
        
        self._gene_to_transcripts: Dict[str, Set[str]] = defaultdict(set)
        self._transcript_to_gene: Dict[str, str] = {}
        
        self._transcript_to_biotype: Dict[str, str] = {}
        self._gene_to_name: Dict[str, str] = {}
        self._name_to_gene: Dict[str, Set[str]] = defaultdict(set)  # One name can map to multiple genes
    
    def add_gene_transcript_mapping(self, gene_id: str, transcript_id: str) -> None:
        # Validate transcript not already mapped to different gene
        if transcript_id in self._transcript_to_gene:
            existing_gene = self._transcript_to_gene[transcript_id]
            if existing_gene != gene_id:
                raise ValueError(
                    f"Transcript {transcript_id} already mapped to {existing_gene}, "
                    f"cannot map to {gene_id}"
                )

        self._gene_to_transcripts[gene_id].add(transcript_id)
        self._transcript_to_gene[transcript_id] = gene_id
    
    def get_transcripts(self, gene_id: str) -> Set[str]:
        return self._gene_to_transcripts.get(gene_id, set()).copy()
    
    def get_gene(self, transcript_id: str) -> Optional[str]:
        return self._transcript_to_gene.get(transcript_id)

    def add_transcript_biotype(self, transcript_id: str, biotype: str) -> None:
        self._transcript_to_biotype[transcript_id] = biotype
    
    def get_transcript_biotype(self, transcript_id: str) -> Optional[str]:
        return self._transcript_to_biotype.get(transcript_id)
    
    # Gene name mapping methods
    def add_gene_name(self, gene_id: str, gene_name: str) -> None:
        self._gene_to_name[gene_id] = gene_name
        self._name_to_gene[gene_name].add(gene_id)
    
    def get_gene_name(self, gene_id: str) -> Optional[str]:
        return self._gene_to_name.get(gene_id)
    
    def get_genes_by_name(self, gene_name: str) -> Set[str]:
        return self._name_to_gene.get(gene_name, set()).copy()

    def has_gene(self, gene_id: str) -> bool:
        return gene_id in self._gene_to_transcripts
    
    def has_transcript(self, transcript_id: str) -> bool:
        return transcript_id in self._transcript_to_gene
    
    def has_gene_transcript_mapping(self) -> bool:
        return bool(self._gene_to_transcripts)
    
    def has_biotype_mapping(self) -> bool:
        return bool(self._transcript_to_biotype)
    
    def has_gene_name_mapping(self) -> bool:
        return bool(self._gene_to_name)

    @property
    def gene_ids(self) -> Set[str]:
        genes = set(self._gene_to_transcripts.keys())
        genes.update(self._gene_to_name.keys())
        return genes
    
    @property
    def transcript_ids(self) -> Set[str]:
        transcripts = set(self._transcript_to_gene.keys())
        transcripts.update(self._transcript_to_biotype.keys())
        return transcripts
    
    def get_gene_transcript_count(self) -> int:
        return len(self._transcript_to_gene)
    
    def get_biotype_count(self) -> int:
        return len(self._transcript_to_biotype)
    
    def get_gene_name_count(self) -> int:
        return len(self._gene_to_name)
    
    def summary(self) -> str:
        mappings = []
        if self.has_gene_transcript_mapping():
            mappings.append(f"gene-transcript: {self.get_gene_transcript_count()} pairs")
        if self.has_biotype_mapping():
            mappings.append(f"transcript-biotype: {self.get_biotype_count()} pairs")
        if self.has_gene_name_mapping():
            mappings.append(f"gene-name: {self.get_gene_name_count()} pairs")
        
        if not mappings:
            return "GeneData: No mappings available"
        
        source_info = f" from {self.source_file}" if self.source_file else ""
        return f"GeneData: {', '.join(mappings)}{source_info}"
    
    def __repr__(self) -> str:
        return self.summary()


"""Test flexible GTF attribute field name parsing.

Tests the enhanced GTF parser's ability to handle different field naming conventions
used by various databases (GENCODE/Ensembl, UCSC, NCBI, etc.).
"""

import pytest
import tempfile
from pathlib import Path
from pyrion.io.gtf import read_gtf


@pytest.mark.io
class TestGTFFlexibleFieldNames:
    """Test GTF parsing with various field naming conventions."""
    
    def create_test_gtf(self, attributes_string):
        """Helper to create a test GTF file with specific attributes."""
        gtf_content = f"""##gff-version 2
chr1	TEST	gene	1000	2000	.	+	.	gene_id "TESTG001"; {attributes_string}
chr1	TEST	transcript	1000	2000	.	+	.	gene_id "TESTG001"; transcript_id "TESTT001"; {attributes_string}
chr1	TEST	exon	1000	1500	.	+	.	gene_id "TESTG001"; transcript_id "TESTT001"; exon_number "1";
chr1	TEST	exon	1600	2000	.	+	.	+	.	gene_id "TESTG001"; transcript_id "TESTT001"; exon_number "2";
"""
        return gtf_content
    
    def test_gencode_standard_naming(self):
        """Test GENCODE/Ensembl standard field names."""
        attributes = 'gene_name "TEST1"; gene_type "protein_coding"; transcript_type "protein_coding";'
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            
            assert len(collection.transcripts) == 1, "Should parse one transcript"
            
            gene_data = collection._gene_data
            assert gene_data is not None, "Should have gene data"
            
            # Test gene name extraction
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name == "TEST1", f"Expected 'TEST1', got '{gene_name}'"
            
            # Test biotype extraction
            biotype = gene_data.get_transcript_biotype("TESTT001")
            assert biotype == "protein_coding", f"Expected 'protein_coding', got '{biotype}'"
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)
    
    def test_ucsc_variant_naming(self):
        """Test UCSC variant field names."""
        attributes = 'geneName "TEST2"; geneType "lncRNA"; transcriptType "lncRNA";'
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            gene_data = collection._gene_data
            
            # Test gene name extraction with UCSC naming
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name == "TEST2", f"Expected 'TEST2', got '{gene_name}'"
            
            # Test biotype extraction with UCSC naming
            biotype = gene_data.get_transcript_biotype("TESTT001")
            assert biotype == "lncRNA", f"Expected 'lncRNA', got '{biotype}'"
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)
    
    def test_ncbi_gff3_style_naming(self):
        """Test NCBI/GFF3 style field names."""
        attributes = 'Name "TEST3"; biotype "pseudogene"; transcript_biotype "pseudogene";'
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            gene_data = collection._gene_data
            
            # Test gene name extraction with GFF3 naming
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name == "TEST3", f"Expected 'TEST3', got '{gene_name}'"
            
            # Test biotype extraction with GFF3 naming  
            biotype = gene_data.get_transcript_biotype("TESTT001")
            assert biotype == "pseudogene", f"Expected 'pseudogene', got '{biotype}'"
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)
    
    def test_alternative_gene_symbol_naming(self):
        """Test alternative gene symbol field names."""
        attributes = 'gene_symbol "TEST4"; gene_biotype "miRNA";'
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            gene_data = collection._gene_data
            
            # Test gene name extraction with gene_symbol
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name == "TEST4", f"Expected 'TEST4', got '{gene_name}'"
            
            # Note: gene_biotype should map to transcript biotype in our current implementation
            # This tests the flexible gene type extraction
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)
    
    def test_mixed_naming_conventions(self):
        """Test GTF with mixed naming conventions."""
        # Mix GENCODE gene_name with UCSC transcriptType
        attributes = 'gene_name "MIXED"; transcriptType "protein_coding";'
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            gene_data = collection._gene_data
            
            # Should handle mixed conventions correctly
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name == "MIXED", f"Expected 'MIXED', got '{gene_name}'"
            
            biotype = gene_data.get_transcript_biotype("TESTT001")
            assert biotype == "protein_coding", f"Expected 'protein_coding', got '{biotype}'"
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)
    
    def test_missing_optional_fields(self):
        """Test GTF parsing when optional fields are missing."""
        # Only required fields, no gene_name or biotype info
        attributes = ''
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            
            # Should still parse successfully
            assert len(collection.transcripts) == 1, "Should parse transcript without optional fields"
            
            gene_data = collection._gene_data
            assert gene_data is not None, "Should have gene data even without optional fields"
            
            # Optional fields should be None/missing
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name is None, f"Expected None for missing gene name, got '{gene_name}'"
            
            # Biotype might still be None if not provided
            biotype = gene_data.get_transcript_biotype("TESTT001")
            # This could be None or extracted from somewhere else - just check it doesn't crash
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)
    
    def test_field_priority_order(self):
        """Test that field extraction follows correct priority order."""
        # Include multiple possible field names - should pick the first match
        attributes = 'Name "SECOND"; gene_name "FIRST"; gene_symbol "THIRD";'
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            gene_data = collection._gene_data
            
            # Should pick gene_name first (appears first in our priority list)
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name == "FIRST", f"Expected 'FIRST' (gene_name priority), got '{gene_name}'"
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)
    
    def test_quoted_values_with_spaces(self):
        """Test handling of quoted values containing spaces."""
        attributes = 'gene_name "Multi Word Gene"; transcript_type "protein coding";'
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            gene_data = collection._gene_data
            
            # Should handle spaces in quoted values
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name == "Multi Word Gene", f"Expected 'Multi Word Gene', got '{gene_name}'"
            
            biotype = gene_data.get_transcript_biotype("TESTT001")
            assert biotype == "protein coding", f"Expected 'protein coding', got '{biotype}'"
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)
    
    def test_special_characters_in_values(self):
        """Test handling of special characters in field values."""
        # Use various special characters that might appear in gene names
        attributes = 'gene_name "GENE-1.2_TEST"; transcript_type "protein_coding";'
        gtf_content = self.create_test_gtf(attributes)
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            collection = read_gtf(tmp_path)
            gene_data = collection._gene_data
            
            # Should handle special characters correctly
            gene_name = gene_data.get_gene_name("TESTG001")
            assert gene_name == "GENE-1.2_TEST", f"Expected 'GENE-1.2_TEST', got '{gene_name}'"
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)


@pytest.mark.io
class TestGTFFlexibleParsingIntegration:
    """Integration tests for flexible GTF parsing with real data."""
    
    @pytest.fixture
    def sample_gtf_file(self):
        """Use the standard test GTF file."""
        return Path("test_data/gtf/sample_gencode.gtf")
    
    def test_real_gtf_flexible_parsing(self, sample_gtf_file):
        """Test that the flexible parsing works with real GTF data."""
        collection = read_gtf(str(sample_gtf_file))
        gene_data = collection._gene_data
        
        # Should have extracted gene names and biotypes
        assert gene_data.has_gene_name_mapping(), "Should extract gene names from real data"
        assert gene_data.has_biotype_mapping(), "Should extract biotypes from real data"
        
        # Test specific known gene from the sample data
        # DDX11L16 should be present
        target_gene_id = "ENSG00000290825.2"
        gene_name = gene_data.get_gene_name(target_gene_id)
        
        assert gene_name is not None, f"Should find gene name for {target_gene_id}"
        assert gene_name == "DDX11L16", f"Expected 'DDX11L16', got '{gene_name}'"
    
    def test_biotype_variety_extraction(self, sample_gtf_file):
        """Test that various biotypes are correctly extracted."""
        collection = read_gtf(str(sample_gtf_file))
        gene_data = collection._gene_data
        
        # Collect all biotypes
        all_biotypes = set()
        for transcript in collection.transcripts:
            biotype = gene_data.get_transcript_biotype(transcript.id)
            if biotype:
                all_biotypes.add(biotype)
        
        # Should have multiple biotype categories
        assert len(all_biotypes) > 1, f"Expected multiple biotypes, got: {all_biotypes}"
        
        # Should include common biotypes
        expected_biotypes = {"protein_coding", "lncRNA"}
        found_expected = all_biotypes.intersection(expected_biotypes)
        assert len(found_expected) > 0, f"Expected to find {expected_biotypes}, got {all_biotypes}"
    
    def test_gene_name_coverage(self, sample_gtf_file):
        """Test gene name extraction coverage."""
        collection = read_gtf(str(sample_gtf_file))
        gene_data = collection._gene_data
        
        # Count genes with names
        genes_with_names = 0
        total_genes = len(gene_data.gene_ids)
        
        for gene_id in gene_data.gene_ids:
            if gene_data.get_gene_name(gene_id) is not None:
                genes_with_names += 1
        
        # Should have reasonable coverage (not all genes may have names)
        coverage = genes_with_names / total_genes if total_genes > 0 else 0
        
        assert coverage > 0, "Should extract some gene names"
        # Don't enforce 100% coverage as it depends on the source data


@pytest.mark.io
class TestGTFFlexibleParsingErrorCases:
    """Test error handling in flexible GTF parsing."""
    
    def test_malformed_attributes(self):
        """Test handling of malformed attribute strings."""
        # Missing quotes, incomplete fields, etc.
        malformed_attributes = [
            'gene_name NOQUOTES; transcript_type "protein_coding";',  # Missing quotes
            'gene_name "UNCLOSED; transcript_type "protein_coding";',  # Unclosed quote
            'gene_name; transcript_type "protein_coding";',  # Missing value
            '"ORPHAN_VALUE"; transcript_type "protein_coding";',  # Orphan value
        ]
        
        for i, attributes in enumerate(malformed_attributes):
            gtf_content = f"""##gff-version 2
chr1	TEST	gene	1000	2000	.	+	.	gene_id "TESTG00{i}"; {attributes}
chr1	TEST	transcript	1000	2000	.	+	.	gene_id "TESTG00{i}"; transcript_id "TESTT00{i}"; {attributes}
chr1	TEST	exon	1000	2000	.	+	.	gene_id "TESTG00{i}"; transcript_id "TESTT00{i}"; exon_number "1";
"""
            
            with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
                tmp.write(gtf_content)
                tmp_path = tmp.name
            
            try:
                # Should not crash, even with malformed attributes
                collection = read_gtf(tmp_path)
                
                # Should still parse the transcript
                assert len(collection.transcripts) == 1, f"Should parse transcript despite malformed attributes (case {i})"
                
                # Gene data should exist but may not extract the malformed fields
                gene_data = collection._gene_data
                assert gene_data is not None, f"Should have gene data (case {i})"
                
            except Exception as e:
                # If it does fail, it should be a reasonable error, not a segfault
                assert "parse" in str(e).lower() or "format" in str(e).lower(), \
                    f"Unexpected error type for malformed case {i}: {e}"
            finally:
                Path(tmp_path).unlink(missing_ok=True)
    
    def test_extremely_long_values(self):
        """Test handling of extremely long field values."""
        long_value = "A" * 10000  # Very long gene name
        attributes = f'gene_name "{long_value}"; transcript_type "protein_coding";'
        
        gtf_content = f"""##gff-version 2
chr1	TEST	gene	1000	2000	.	+	.	gene_id "TESTG001"; {attributes}
chr1	TEST	transcript	1000	2000	.	+	.	gene_id "TESTG001"; transcript_id "TESTT001"; {attributes}
chr1	TEST	exon	1000	2000	.	+	.	gene_id "TESTG001"; transcript_id "TESTT001"; exon_number "1";
"""
        
        with tempfile.NamedTemporaryFile(mode='w', suffix='.gtf', delete=False) as tmp:
            tmp.write(gtf_content)
            tmp_path = tmp.name
        
        try:
            # Should handle long values gracefully
            collection = read_gtf(tmp_path)
            gene_data = collection._gene_data
            
            gene_name = gene_data.get_gene_name("TESTG001")
            
            # Should either extract the full long name or handle it gracefully
            if gene_name is not None:
                assert len(gene_name) > 1000, "Should extract long gene name"
            
        finally:
            Path(tmp_path).unlink(missing_ok=True)


"""Gene data I/O support."""

from typing import Union, Optional, List, Dict
from pathlib import Path

from ..core.gene_data import GeneData


def resolve_index(column_idx: Union[int, str], header: List[str]) -> int | None:
    if column_idx is None:
        return None

    if isinstance(column_idx, int):
        return column_idx - 1

    if isinstance(column_idx, str):
        try:
            return header.index(column_idx)
        except ValueError:
            raise ValueError(f"Column '{column_idx}' not found in header")


def read_gene_data(
    file_path: Union[str, Path],
    gene_column: Optional[Union[int, str]] = None,
    transcript_id_column: Optional[Union[int, str]] = None,
    gene_name_column: Optional[Union[int, str]] = None,
    transcript_type_column: Optional[Union[int, str]] = None,
    separator: str = '\t',
    has_header: bool = True,
) -> GeneData:
    """Read gene data from TSV/CSV file and build mappings."""
    file_path = Path(file_path)

    gene_data = GeneData(source_file=str(file_path))
    with open(file_path, 'r') as f:
        lines = f.readlines()

    if has_header:
        header = lines[0].strip().split(separator)
        data_lines = lines[1:]
    else:
        header = None
        data_lines = lines

    transcript_column_idx = resolve_index(transcript_id_column, header)
    gene_column_idx = resolve_index(gene_column, header)
    gene_name_column_idx = resolve_index(gene_name_column, header)
    transcript_type_column_idx = resolve_index(transcript_type_column, header)

    fill_gene_to_transcripts = gene_column_idx is not None and transcript_column_idx is not None
    fill_gene_to_name = gene_name_column_idx is not None and gene_column_idx is not None
    fill_transcript_to_biotype = transcript_type_column_idx is not None and transcript_column_idx is not None

    for line in data_lines:
        line = line.strip().split(separator)
        transcript_id = line[transcript_column_idx]
        gene_id = line[gene_column_idx] if gene_column_idx is not None else None
        gene_name = line[gene_name_column_idx] if gene_name_column_idx is not None else None
        transcript_type = line[transcript_type_column_idx] if transcript_type_column_idx is not None else None

        if fill_gene_to_transcripts:
            gene_data.add_gene_transcript_mapping(gene_id, transcript_id)
        if fill_gene_to_name:
            gene_data.add_gene_name(gene_id, gene_name)
        if fill_transcript_to_biotype:
            gene_data.add_transcript_biotype(transcript_id, transcript_type)

    return gene_data


def write_gene_data_tsv(
    gene_data: GeneData,
    file_path: Union[str, Path],
    include_gene_transcript: bool = True,
    include_transcript_biotype: bool = True,
    include_gene_name: bool = True,
    separator: str = '\t'
) -> None:
    file_path = Path(file_path)
    has_gene_transcript = gene_data.has_gene_transcript_mapping() and include_gene_transcript
    has_transcript_biotype = gene_data.has_biotype_mapping() and include_transcript_biotype
    has_gene_name = gene_data.has_gene_name_mapping() and include_gene_name
    
    if not any([has_gene_transcript, has_transcript_biotype, has_gene_name]):
        raise ValueError("No data mappings available to export")
    
    header = ["transcript_id"]
    if has_gene_transcript:
        header.append("gene_id")
    if has_transcript_biotype:
        header.append("transcript_biotype")
    if has_gene_name:
        header.append("gene_name")
    
    transcript_ids = set()
    if has_gene_transcript:
        transcript_ids.update(gene_data.transcript_ids)
    if has_transcript_biotype:
        transcript_ids.update(gene_data.transcript_ids)
    
    if not transcript_ids:
        raise ValueError("No transcript data available to export")
    
    with open(file_path, 'w') as f:
        f.write(separator.join(header) + '\n')

        for transcript_id in sorted(transcript_ids):
            row = [transcript_id]
            
            if has_gene_transcript:
                gene_id = gene_data.get_gene(transcript_id) or ""
                row.append(gene_id)
            
            if has_transcript_biotype:
                biotype = gene_data.get_transcript_biotype(transcript_id) or ""
                row.append(biotype)
            
            if has_gene_name:
                gene_id = gene_data.get_gene(transcript_id) if has_gene_transcript else None
                gene_name = ""
                if gene_id:
                    gene_name = gene_data.get_gene_name(gene_id) or ""
                row.append(gene_name)
            
            f.write(separator.join(row) + '\n')
